##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HSDS (HDF5 Scalable Data Service), Libraries and      #
# Utilities.  The full HSDS copyright notice, including                      #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

import os
import sys
import logging
import os.path as op

try:
    import h5py
    import h5pyd
except ImportError as e:
    sys.stderr.write(f"ERROR : {e} : install it to use this utility...\n")
    sys.exit(1)

try:
    import s3fs

    S3FS_IMPORT = True
except ImportError:
    S3FS_IMPORT = False

if __name__ == "__main__":
    from config import Config
    from utillib import load_file
else:
    from .config import Config
    from .utillib import load_file

from urllib.parse import urlparse

cfg = Config()

# ----------------------------------------------------------------------------------
def abort(msg):
    logging.error(msg)
    if cfg["logfile"]:
        sys.stderr.write(msg + "\n")
    logging.error("exiting program with return code -1")
    sys.exit(-1)

# ----------------------------------------------------------------------------------
def usage():
    option_names = cfg.get_names()
    cmd = cfg.get_cmd()
    print("Usage:\n")
    print(f"    {cmd} [ OPTIONS ]  sourcefile  domain")
    print(f"    {cmd} [ OPTIONS ]  sourcefile_1, sourcefile_2,...  folder")
    print("")
    print("Description:")
    print("    Copy HDF5 file to domain or multiple files to a domain folder")
    print("       sourcefile: HDF5 file to be copied ")
    print("       domain: HSDS domain (absolute path with or without hdf5:// prefix)")
    print("       folder: HSDS folder (path as above ending in '/')")
    print("")
    print("Options:")
    for name in option_names:
        help_msg = cfg.get_help_message(name)
        if help_msg:
            print(f"    {help_msg}")  
    print("")
    print("Note about --link option:")
    print("    --link enables just the source HDF5 metadata to be ingested while the dataset data")
    print("     is left in the original file and fetched as needed.")
    print("     When used with files stored in AWS S3, the source file can be specified using the S3")
    print("     path: 's3://<bucket_name>/<s3_path>'.  Preferably, the bucket should be in the same")
    print("     region as the HSDS service")
    print("     For Posix or Azure deployments, the source file needs to be copied to a regular file,")
    print("     and the --linkpath option should be used to specifiy the Azure container name and path, or ")
    print("     (for HSDS deployed with POSIX) the file path relative to the server ROOT_DIR")
    print("")
    print(cfg.get_see_also(cmd))
    print("")
    sys.exit(-1)
     


# end print_usage


# ----------------------------------------------------------------------------------
def main():

    COMPRESSION_FILTERS = ("blosclz", "lz4", "lz4hc", "snappy", "gzip", "zstd")

    s3 = None # S3FS instance

    cfg.setitem("append", False, flags=["-a", "--append"], help="append to existing domain")
    cfg.setitem("extend_dim", None, flags=["--extend",], choices=["DIMSCALE",], help="extend along given dimensionscale")
    cfg.setitem("extend_offset", None, flags=["--extend-offset"], choices=["N",], help="write data at index n along extended dimension")
    cfg.setitem("no_clobber", False, flags=["-n", "--no-clobber"],  help="do not overwrite target")
    cfg.setitem("nodata", False, flags=["--nodata",], help="do not copy dataset data")
    cfg.setitem("z", None, flags=["-z",], choices=["N",], help="apply compression filter to any non-compressed datasets, n: [0-9]")
    cfg.setitem("link", None, flags=["--link",],  help="Link to dataset data (sourcefile given as <bucket>/<path>) or s3uri")
    cfg.setitem("fastlink", None, flags=["--fastlink",],  help="Link to dataset data without initializing chunk locations (will be set server side)")
    cfg.setitem("linkpath", None, flags=["--linkpath",], choices=["PATH_URI",], help="Use the given URI for the link references rather than the src path")
    cfg.setitem("compression", None, flags=["--compression",], choices=COMPRESSION_FILTERS, help="use the given compression algorithm for -z option (lz4 is default)")
    cfg.setitem("ignorefilters", False, flags=["--ignore-filters"], help="ignore any filters used by source dataset")
    cfg.setitem("retries", 3, flags=["--retries",], choices=["N",], help="Set number of server retry attempts")
    cfg.setitem("help", False, flags=["-h", "--help"], help="this message")

    try:
        cmdline_values = cfg.set_cmd_flags(sys.argv[1:])
    except ValueError as ve:
        print(ve)
        usage()

    if len(cmdline_values) < 2:
        usage()

    domain = cmdline_values[-1]
    src_files = cmdline_values[:-1]

    # setup logging
    logfname = cfg["logfile"]
    loglevel = cfg.get_loglevel()
    logging.basicConfig(filename=logfname, format='%(levelname)s %(asctime)s %(message)s', level=loglevel)
    logging.debug(f"set log_level to {loglevel}") 

    if cfg["linkpath"] and not cfg["link"]:
        abort("--linkpath option can only be used with --link")

    if cfg["extend_offset"] and cfg["extend_dim"] is None:
        abort("--extend-offset option can only be used with --link")

    if cfg["extend_dim"] and cfg["link"]:
        abort("--extend option can't be used with --link")

    if cfg["nodata"] and cfg["link"]:
        abort("--nodata option can't  be used with --link")

    if cfg["link"]:
        dataload = "link"
    elif cfg["fastlink"]:
        dataload = "fastlink"
    elif cfg["nodata"]:
        dataload = None
    else:
        dataload = "ingest"
    
    logging.info(f"source files: {src_files}")
    logging.info(f"target domain: {domain}")
    if len(src_files) > 1 and domain[-1] != "/":
        abort("target must be a folder if multiple source files are provided")

    # check we have min HDF5 lib version for link option
    if cfg["link"]:
        logging.info("checking libversion")

        if (
            h5py.version.version_tuple.major == 2
            and h5py.version.version_tuple.minor < 10
        ):
            abort("link option requires h5py version 2.10 or higher")
            
        if h5py.version.hdf5_version_tuple < (1, 10, 6):
            abort("link option requires h5py version 2.10 or higher")
        

    try:

        for src_file in src_files:
            # check if this is a non local file, if it is remote (http, etc...) stage it first then insert it into hsds
            src_file_chk = urlparse(src_file)
            logging.debug(src_file_chk)

            tgt = domain
            if tgt[-1] == "/":
                # folder destination
                tgt = tgt + op.basename(src_file)

            # get a handle to input file
            if src_file.startswith("s3://"):
                s3path = src_file
                if not S3FS_IMPORT:
                    abort("Install S3FS package to load s3 files")
                    sys.exit(1)

                if not s3:
                    key = os.environ.get("AWS_ACCESS_KEY_ID")
                    secret = os.environ.get("AWS_SECRET_ACCESS_KEY")
                    aws_s3_gateway = os.environ.get("AWS_GATEWAY")
                    client_kwargs = {}
                    if aws_s3_gateway:
                        client_kwargs["endpoint_url"] = aws_s3_gateway

                    s3 = s3fs.S3FileSystem(use_ssl=False, key=key, secret=secret, client_kwargs=client_kwargs)
                try:
                    fin = h5py.File(s3.open(src_file, "rb"), moe="r")
                except IOError as ioe:
                    abort(f"Error opening file {src_file}: {ioe}")
                    
            else:
                if cfg["link"] or cfg["fastlink"]:
                    if op.isabs(src_file) and not cfg["linkpath"]:
                        msg = "source file must be an s3path (for HSDS using S3 storage) or relative path from server "
                        msg += "root directory (for HSDS using posix storage)"
                        abort(msg)
                    s3path = src_file
                else:
                    s3path = None
                try:
                    fin = h5py.File(src_file, mode="r")
                except IOError as ioe:
                    abort(f"Error opening file {src_file}: {ioe}")

            # create the output domain
            try:
                if cfg["append"]:
                    mode = "a"
                elif cfg["no_clobber"]:
                    mode = "x"
                else:
                    mode = "w"
                kwargs = {
                    "username": cfg["hs_username"],
                    "password": cfg["hs_password"],
                    "endpoint": cfg["hs_endpoint"],
                    "bucket": cfg["hs_bucket"],
                    "mode": mode,
                    "retries": cfg["retries"],
                }

                fout = h5pyd.File(tgt, **kwargs)
            except IOError as ioe:
                if ioe.errno == 404:
                    abort(f"Domain: {tgt} not found")
                elif ioe.errno == 403:
                    abort(f"No write access to domain: {tgt}")
                else:
                    abort(f"Error creating file {tgt}: {ioe}")

            if cfg["linkpath"]:
                # now that we have a handle to the source file,
                # repurpose s3path to the s3uri that will actually get stored 
                # in the target domain
                s3path = cfg["linkpath"]

            if cfg["no_clobber"]:
                if cfg["append"]: 
                    # no need to check for clobber if not in append mode
                    no_clobber = True
                else:
                    no_clobber = False
            else:
                no_clobber = False

            if cfg["compression"]:
                compression = cfg["compression"]
            else:
                compression = None

            if cfg["z"]:
                try:
                    compression_opts = int(cfg["z"])
                    if compression is None:
                        # if no other comressor is specified, just use gzip
                        compression = "gzip"

                except ValueError:
                    # not a numeric option?  Just pass the string
                    compression_opts = cfg["z"]
            else:
                compression_opts = None


            # do the actual load
            kwargs = {
                "verbose": cfg["verbose"],
                "dataload": dataload,
                "s3path": s3path,
                "compression": compression,
                "compression_opts": compression_opts,
                "ignorefilters": cfg["ignorefilters"],
                "append": cfg["append"],
                "extend_dim": cfg["extend_dim"],
                "extend_offset": cfg["extend_offset"],
                "ignore_error": cfg["ignore_error"],
                "no_clobber": no_clobber
                            }
            load_file(fin, fout, **kwargs)

            msg = f"File {src_file} uploaded to domain: {tgt}"
            logging.info(msg)
            if cfg["verbose"]:
                print(msg)

    except KeyboardInterrupt:
        abort("Aborted by user via keyboard interrupt.")


# __main__
if __name__ == "__main__":
    main()



# This file is part of h5py, a Python interface to the HDF5 library.
#
# http://www.h5py.org
#
# Copyright 2008-2013 Andrew Collette and contributors
#
# License:  Standard 3-clause BSD; see "license.txt" for full license terms
#           and contributor agreement.

"""
    Implements high-level access to committed datatypes in the file. """

from __future__ import absolute_import

import posixpath as pp

#from ..h5t import TypeID
from .base import HLObject

from .objectid import  TypeID
from .h5type import createDataType

class Datatype(HLObject):

    """
        Represents an HDF5 named datatype stored in a file.

        To store a datatype, simply assign it to a name in a group:

        >>> MyGroup["name"] = numpy.dtype("f")
        >>> named_type = MyGroup["name"]
        >>> assert named_type.dtype == numpy.dtype("f")  """

    @property
    def dtype(self):
        """Numpy dtype equivalent for this datatype"""
        return self._dtype

    def __init__(self, bind):
        """ Create a new Datatype object by binding to a low-level TypeID.
        """
        if not isinstance(bind, TypeID):
            # todo: distinguish type from other hl objects
            raise ValueError("%s is not a TypeID" % bind)
        HLObject.__init__(self, bind)

        self._dtype = createDataType(self.id.type_json)
        self._req_prefix = "/datatypes/" + self.id.uuid


    def __repr__(self):
        if not self.id:
            return "<Closed HDF5 named type>"
        if self.name is None:
            namestr = '("anonymous")'
        else:
            name = pp.basename(pp.normpath(self.name))
            namestr = '"%s"' % (name if name != '' else '/')
            if name:
                namestr = f'"{name}"'
            else:
                namestr = '"/"'
        return f'<HDF5 named type {namestr} (dtype {self.dtype.str})>'



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

"""
    Implements high-level operations for attributes.

    Provides the AttributeManager class, available on high-level objects
    as <obj>.attrs.
"""

from __future__ import absolute_import

import base64
import numpy
import json
import numpy as np

from . import base
from .base import jsonToArray, Empty
from .datatype import Datatype
from .objectid import GroupID, DatasetID, TypeID
from .h5type import getTypeItem, createDataType, special_dtype, Reference


class AttributeManager(base.MutableMappingHDF5, base.CommonStateObject):

    """
        Allows dictionary-style access to an HDF5 object's attributes.

        These are created exclusively by the library and are available as
        a Python attribute at <object>.attrs

        Like Group objects, attributes provide a minimal dictionary-
        style interface.  Anything which can be reasonably converted to a
        Numpy array or Numpy scalar can be stored.

        Attributes are automatically created on assignment with the
        syntax <obj>.attrs[name] = value, with the HDF5 type automatically
        deduced from the value.  Existing attributes are overwritten.

        To modify an existing attribute while preserving its type, use the
        method modify().  To specify an attribute of a particular type and
        shape, use create().
    """

    def __init__(self, parent):
        """ Private constructor.
        """
        self._parent = parent

        if isinstance(parent.id, GroupID):
            self._req_prefix = "/groups/" + parent.id.uuid + "/attributes/"
        elif isinstance(parent.id, TypeID):
            self._req_prefix = "/datatypes/" + parent.id.uuid + "/attributes/"
        elif isinstance(parent.id, DatasetID):
            self._req_prefix = "/datasets/" + parent.id.uuid + "/attributes/"
        else:
            # "unknown id"
            self._req_prefix = "<unknown>"
        objdb = self._parent.id.http_conn.getObjDb()
        if objdb:
            # _objdb is meta-data pulled from the domain on open.
            # see if we can extract the link json from there
            objid = self._parent.id.uuid
            if objid not in objdb:
                raise IOError("Expected to find {} in objdb".format(objid))
            obj_json = objdb[objid]
            self._objdb_attributes = obj_json["attributes"]
        else:
            self._objdb_attributes = None

    def _bytesArrayToList(self, data):
        """
        Convert list that may contain bytes type elements to list of string
        elements
        """
        text_types = (bytes, str)
        if isinstance(data, text_types):
            is_list = False
        elif isinstance(data, (numpy.ndarray, numpy.generic)):
            if len(data.shape) == 0:
                is_list = False
                data = data.tolist()  # tolist will return a scalar in this case
                if type(data) in (list, tuple):
                    is_list = True
                else:
                    is_list = False
            else:
                is_list = True
        elif isinstance(data, list) or isinstance(data, tuple):
            is_list = True
        else:
            is_list = False

        if is_list:
            out = []
            for item in data:
                out.append(self._bytesArrayToList(item)) # recursive call
        elif isinstance(data, bytes):
            print("got data bytes:", data)
            try:
                out = data.decode("utf-8")
            except UnicodeDecodeError:
                print("unicode decode error")
        else:
            out = data

        return out

    def __getitem__(self, name):
        """ Read the value of an attribute.
        """
        if isinstance(name, bytes):
            name = name.decode("utf-8")

        if self._objdb_attributes is not None:
            if name not in self._objdb_attributes:
                raise KeyError
            attr_json = self._objdb_attributes[name]
        else:
            req = self._req_prefix + name
            try:
                attr_json = self._parent.GET(req)
            except IOError:
                raise KeyError

        shape_json = attr_json['shape']
        type_json = attr_json['type']
        dtype = createDataType(type_json)
        if shape_json['class'] == 'H5S_NULL':
            return Empty(dtype)
        if "value" in attr_json:
            value_json = attr_json['value']
        else:
            print("attr_json:", attr_json)
            value_json = None

        if 'dims' in shape_json:
            shape = shape_json['dims']
        else:
            shape = ()

        # Do this first, as we'll be fiddling with the dtype for top-level
        # array types
        htype = dtype

        # NumPy doesn't support top-level array types, so we have to "fake"
        # the correct type and shape for the array.  For example, consider
        # attr.shape == (5,) and attr.dtype == '(3,)f'. Then:
        if dtype.subdtype is not None:
            subdtype, subshape = dtype.subdtype
            shape = shape + subshape   # (5, 3)
            dtype = subdtype           # 'f'

        arr = jsonToArray(shape, htype, value_json)

        if len(arr.shape) == 0:
            v = arr[()]
            if isinstance(v, str):
                # if this is not utf-8, return bytes instead
                try:
                    v.encode("utf-8")
                except UnicodeEncodeError:
                    self._parent.log.debug("converting utf8 unencodable string as bytes")
                    v = v.encode("utf-8", errors="surrogateescape")
            return v

        return arr

    def __setitem__(self, name, value):
        """ Set a new attribute, overwriting any existing attribute.

        The type and shape of the attribute are determined from the data.  To
        use a specific type or shape, or to preserve the type of an attribute,
        use the methods create() and modify().
        """
        self.create(name, data=value, dtype=base.guess_dtype(value))

    def __delitem__(self, name):
        """ Delete an attribute (which must already exist). """
        if isinstance(name, bytes):
            name = name.decode("utf-8")
        req = self._req_prefix + name
        self._parent.DELETE(req)

    def create(self, name, data, shape=None, dtype=None):
        """ Create a new attribute, overwriting any existing attribute.

        name
            Name of the new attribute (required)
        data
            An array to initialize the attribute (required)
        shape
            Shape of the attribute.  Overrides data.shape if both are
            given, in which case the total number of points must be unchanged.
        dtype
            Data type of the attribute.  Overrides data.dtype if both
            are given.
        """
        self._parent.log.info("attrs.create({})".format(name))

        # First, make sure we have a NumPy array.  We leave the data
        # type conversion for HDF5 to perform.
        if isinstance(data, Reference):
            dtype = special_dtype(ref=Reference)
        if not isinstance(data, Empty):
            data = numpy.asarray(data, dtype=dtype, order='C')

        if shape is None and not isinstance(data, Empty):
            shape = data.shape

        use_htype = None    # If a committed type is given, we must use it
                            # in the call to h5a.create.

        if isinstance(dtype, Datatype):
            use_htype = dtype.id
            dtype = dtype.dtype

            # Special case if data are complex numbers
            if (data.dtype.kind == 'c' and
                (dtype.names is None or
                    dtype.names != ('r', 'i') or
                    any(dt.kind != 'f' for dt, off in dtype.fields.values()) or
                    dtype.fields['r'][0] == dtype.fields['i'][0])):
                raise TypeError(
                    'Wrong committed datatype for complex numbers: %s' %
                    dtype.name)
        elif dtype is None:
            if data.dtype.kind == 'U':
                # use vlen for unicode strings
                dtype = special_dtype(vlen=str)
            else:
                dtype = data.dtype
        else:
            dtype = numpy.dtype(dtype) # In case a string, e.g. 'i8' is passed

        # Where a top-level array type is requested, we have to do some
        # fiddling around to present the data as a smaller array of
        # subarrays.
        if not isinstance(data, Empty):
            if dtype.subdtype is not None:

                subdtype, subshape = dtype.subdtype

                # Make sure the subshape matches the last N axes' sizes.
                if shape[-len(subshape):] != subshape:
                    raise ValueError("Array dtype shape %s is incompatible with data shape %s" % (subshape, shape))

                # New "advertised" shape and dtype
                shape = shape[0:len(shape) - len(subshape)]
                dtype = subdtype

            # Not an array type; make sure to check the number of elements
            # is compatible, and reshape if needed.
            else:
                if numpy.prod(shape) != numpy.prod(data.shape):
                    raise ValueError("Shape of new attribute conflicts with shape of data")

                if shape != data.shape:
                    data = data.reshape(shape)

            # We need this to handle special string types.
        
                data = numpy.asarray(data, dtype=dtype)

        # Make HDF5 datatype and dataspace for the H5A calls
        if use_htype is None:
            type_json = getTypeItem(dtype)
            self._parent.log.debug("attrs.create type_json: {}".format(type_json))

        # This mess exists because you can't overwrite attributes in HDF5.
        # So we write to a temporary attribute first, and then rename.

        req = self._req_prefix + name
        body = {}
        body['type'] = type_json
        print("type_json:", type_json)
        if isinstance(data, Empty):
            body['shape'] = 'H5S_NULL'
        else:
            body['shape'] = shape
            if data.dtype.kind != 'c':
                body_value = self._bytesArrayToList(data)
                print("attr_name:", name)
                print("===============")
                print("attr value:", body_value)
                print("attr type:", type(body_value))
                print("data:", data)
                print("data type:", type(data))
                print("data dtype:", type(data.dtype))
                print("data bytes:", data.tobytes())
                print("data shape:", data.shape)
                print("count:", np.prod(data.shape))
                if np.prod(data.shape) == 1:
                    print("scalar value:", data[()])
                if isinstance(body_value, str):
                    # verify that it is utf8 encodable
                    encodable = False
                    try:
                        data_bytes = body_value.encode("utf8")
                        print("got data_bytes:", data_bytes)
                        encodable = True
                    except UnicodeEncodeError:
                        print("can't encode data")
                    if not encodable:
                        server_version = self._parent.id.http_conn.server_version()
                        print("server_version:", server_version)
                        if server_version and server_version.startswith("0.9"):
                            print('encode data')
                            data_bytes = data.tobytes()
                            body_value = base64.b64encode(data_bytes).decode("ascii")
                            body["encoding"] = "base64"
                            type_json["length"] = len(data_bytes)  # convert to fixed length
                            body["type"] = type_json
                print("setting body value to:", body_value)
                body["value"] = body_value
            else:
                # Special case: complex numbers
                special_dt = createDataType(type_json)
                tmp = numpy.empty(shape=data.shape, dtype=special_dt)
                tmp['r'] = data.real
                tmp['i'] = data.imag
                body['value'] = json.loads(json.dumps(tmp.tolist()))

        try:
            print("attribute put:", body)
            self._parent.PUT(req, body=body)
        except RuntimeError:
            print("runtime error")
            # Resource already exist, try deleting it
            self._parent.log.info("Update to existing attribute ({}), deleting it".format(name))
            self._parent.DELETE(req)
            # now add again
            self._parent.PUT(req, body=body)
        except Exception as e:
            print(f"exception: {e}")

    def modify(self, name, value):
        """ Change the value of an attribute while preserving its type.

        Differs from __setitem__ in that if the attribute already exists, its
        type is preserved.  This can be very useful for interacting with
        externally generated files.

        If the attribute doesn't exist, it will be automatically created.
        """
        pass
        # TBD
        """
        with phil:
            if not name in self:
                self[name] = value
            else:
                value = numpy.asarray(value, order='C')

                attr = h5a.open(self._id, self._e(name))

                if attr.get_space().get_simple_extent_type() == h5s.NULL:
                    raise IOError("Empty attributes can't be modified")

                # Allow the case of () <-> (1,)
                if (value.shape != attr.shape) and not \
                   (numpy.prod(value.shape) == 1 and numpy.prod(attr.shape) == 1):
                    raise TypeError("Shape of data is incompatible with existing attribute")
                attr.write(value)
        """

    def __len__(self):
        """ Number of attributes attached to the object. """

        if self._objdb_attributes is not None:
            count = len(self._objdb_attributes)
        else:
            # make a server requests
            req = self._req_prefix
            # backup over the '/attributes/' part of the req
            req = req[:-(len('/attributes/'))]
            rsp = self._parent.GET(req)  # get parent obj
            count = rsp['attributeCount']
        return count

    def __iter__(self):
        """ Iterate over the names of attributes. """
        if self._objdb_attributes is not None:

            for name in self._objdb_attributes:
                yield name

        else:
            # make server request
            req = self._req_prefix
            # backup over the trailing slash in req
            req = req[:-1]
            rsp = self._parent.GET(req)
            attributes = rsp['attributes']

            attrlist = []
            for attr in attributes:
                attrlist.append(attr['name'])

            for name in attrlist:
                yield name

    def __contains__(self, name):
        """ Determine if an attribute exists, by name. """
        exists = True
        if isinstance(name, bytes):
            name = name.decode("utf-8")

        if self._objdb_attributes is not None:
            exists = name in self._objdb_attributes
        else:
            # make server request
            req = self._req_prefix + name
            try:
                self._parent.GET(req)
            except IOError:
                #todo - verify this is a 404 response
                exists = False
        return exists

    def __repr__(self):
        if not self._parent.id.id:
            return "<Attributes of closed HDF5 object>"
        return "<Attributes of HDF5 object at %s>" % id(self._parent.id)



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

import sys
import logging
import h5pyd
if __name__ == "__main__":
    from config import Config
else:
    from .config import Config

cfg = Config()

#
# log error and abort app
#
def abort(msg):
    logging.error(msg)
    if cfg["logfile"]:
        # write to stderr if we are output logs to a file
        sys.stderr.write(msg + "\n")
    logging.error("exiting program with return code -1")
    sys.exit(-1)

#
# get given ACL, return None if not found
#
def getACL(f, username="default"):
    try:
       acl = f.getACL(username)
    except IOError as ioe:
        if ioe.errno == 403:
            print("No permission to read ACL for this domain")
            sys.exit(1)
        elif ioe.errno == 401:
            print("username/password needs to be provided")
            sys.exit(1)
        elif ioe.errno == 404 or not ioe.errno:
            return None
        else:
            print("unexpected error: {}".format(ioe))
            sys.exit(1)
    if acl and "domain" in acl:
        # remove the domain key
        del acl["domain"]
    return acl

#
# Usage
#
def usage():
    option_names = cfg.get_names()
    cmd = cfg.get_cmd()
    print("Usage:")
    print("")
    print(f"    {cmd} [options] domain [+crudep] [-crudep] [userid1 userid2 ...]")
    print("")
    print("Description:")
    print("    Display, add, or change ACLs for a domain or folder")
    print("")
    print("Options:")
    for name in option_names:
        help_msg = cfg.get_help_message(name)
        if help_msg:
            print(f"    {help_msg}")  
    print("")

    print("Arguments:")
    print("    domain :: Domain or Folder to be updated")
    print("    +/- :: add or remove permissions")
    print("    crudep :: permission flags: Create, Read, Update, Delete, rEadacl, uPdateacl")
    msg = "    userid1, userid2, etc.: list of usernames, group names (group names distinguished "
    msg += "by 'g:' prefix) or 'default' to set permissions for those not otherwise listed"
    print(msg)
    print("")
    print("Examples:")
    print(f"    list acls: {cmd} /home/jill/myfile.h5")
    print(f"    list ted's acl (if any): {cmd} /home/jill/myfile.h5  ted")
    print(f"    add/update acl to give ted read & update permissions: {cmd} /home/jill/myfile.h5 +ru ted")
    print(f"    remove all permissions except read for jill: {cmd} /home/jill/myfile.h5 -cudep jill")
    print(f"    enable create, update, and read ACL for devs group: {cmd} /shared/datafile.h5 +cup g:devs")
    print(f"    enable domain and ACLs to be read by anyone: {cmd} /home/jill/myfile.h5 +re default")
    print("")
    print(cfg.get_see_also(cmd))
    print("")
    sys.exit()



def main():
    perm_abvr = {'c':'create', 'r': 'read', 'u': 'update', 'd': 'delete', 'e': 'readACL', 'p':'updateACL'}
    fields = ('username', 'create', 'read', 'update', 'delete', 'readACL', 'updateACL')
    domain = None
    perm = None
    usernames = []
    add_list = set()
    remove_list = set()

    # additional options
    cfg.setitem("help", False, flags=["-h", "--help"], help="this message")

    
    try:
        cmdline_args = cfg.set_cmd_flags(sys.argv[1:], allow_post_flags=True)
    except ValueError as ve:
        print(ve)
        usage()
    
    if len(cmdline_args) == 0:
        # need a domain
        usage()

    # setup logging
    logfname = cfg["logfile"]
    loglevel = cfg.get_loglevel()
    logging.basicConfig(filename=logfname, format='%(levelname)s %(asctime)s %(message)s', level=loglevel)
    logging.debug(f"set log_level to {loglevel}")

    for arg in cmdline_args:
        if domain is None:
            domain = arg
        elif arg[0] == '+':
            if len(usernames) > 0:
                abort("no usernames given!")
            add_list = set(arg[1:])
        elif arg[0] == '-':
            if len(usernames) > 0:
                abort("remove flags must be placed before usernames!")
            remove_list = set(arg[1:])
        else:
            if arg.find('/') >= 0:
                abort(f"invalid username: {arg}")
            usernames.append(arg)

    logging.info(f"domain: {domain}")
    logging.info(f"add_list: {add_list}")
    logging.info(f"remove_list: {remove_list}")
    logging.info(f"usernames: {usernames}")

    if len(usernames) == 0 and (add_list or remove_list):
        abort("at least one username must be given to add/remove permissions")

    if domain is None:
        abort("no domain specified")

    conflicts = list(add_list & remove_list)

    if len(conflicts) > 0:
        abort(f"permission: {conflicts[0]} flag set for both add and remove")

    mode = 'r'
    if add_list or remove_list:
        mode = 'a'  # we'll be updating the domain
        perm = {}
        for x in add_list:
            if x not in perm_abvr:
                abort("Permission flag: {x} is not valid - must be one of 'crudep'")
            perm_name = perm_abvr[x]
            perm[perm_name] = True
        for x in remove_list:
            if x not in perm_abvr:
                abort(f"Permission flag: {x} is not valid - must be one of 'crudep'")
            perm_name = perm_abvr[x]
            perm[perm_name] = False
        logging.info("perm:", perm)

    # open the domain or folder
    try:
        if domain[-1] == '/':
            f = h5pyd.Folder(domain, mode=mode, endpoint=cfg["hs_endpoint"], username=cfg["hs_username"], password=cfg["hs_password"], bucket=cfg["hs_bucket"])
        else:
            f = h5pyd.File(domain, mode=mode, endpoint=cfg["hs_endpoint"], username=cfg["hs_username"], password=cfg["hs_password"], bucket=cfg["hs_bucket"])
    except IOError as ioe:
        if ioe.errno in (404, 410):
            abort("domain not found")
        elif ioe.errno in (401, 403):
            abort("access is not authorized")
        else:
            abort(f"Unexpected error: {ioe}")

    # update/add ACL if permission flags have been set
    if perm:
        default_acl = {'updateACL': False,
                       'delete': False,
                       'create': False,
                       'read': False,
                       'update': False,
                       'readACL': False,
                       'userName': 'default'
                       }
        # note: list.copy not supported in py2.7, copy by hand for now
        # update_names = usernames.copy()
        update_names = []
        for username in usernames:
            update_names.append(username)

        if not update_names:
            update_names.append("default")

        for username in update_names:
            # get user's ACL if it exist
            acl = getACL(f, username=username)
            if acl is None:
                acl = default_acl.copy()
            acl["userName"] = username
            logging.info(f"updating acl to: {acl}")
            # mix in any permission changes
            for k in perm:
                acl[k] = perm[k]
            try:
                f.putACL(acl)
            except IOError as ioe:
                if ioe.errno in (401, 403):
                    abort("access is not authorized")
                else:
                    abort("Unexpected error:", ioe)
    #
    # read the acls
    #
    if len(usernames) == 0:
        # no usernames, dump all ACLs
        try:
            acls = f.getACLs()
        except IOError as ioe:
            if ioe.errno == 403:
                username = cfg["hs_username"]
                abort(f"User: {username} does not have permission to read ACL for this domain")
            elif ioe.errno == 401:
                abort("username/password needs to be provided")
            else:
                abort(f"Unexpected error: {ioe}")
        print("%015s   %08s  %08s  %08s  %08s  %08s  %08s " % fields)
        print("-"*80)
        for acl in acls:
            vals = (acl["userName"], acl["create"], acl["read"], acl["update"], acl["delete"], acl["readACL"], acl["updateACL"])
            print("%015s   %08s  %08s  %08s  %08s  %08s  %08s " % vals)
    else:
        header_printed = False  # don't print header until we have at least one ACL
        for username in usernames:
            try:
                acl = f.getACL(username)
                if not header_printed:
                    print("%015s   %08s  %08s  %08s  %08s  %08s  %08s " % fields)
                    print("-"*80)
                    header_printed = True
                vals = (acl["userName"], acl["create"], acl["read"], acl["update"], acl["delete"], acl["readACL"], acl["updateACL"])
                print("%015s   %08s  %08s  %08s  %08s  %08s  %08s " % vals)
            except IOError as ioe:
                if ioe.errno == 403:
                    this_user = cfg["hs_username"]
                    abort(f"User {this_user} does not have permission to read ACL for this domain")
                elif ioe.errno == 401:
                    abort("username/password needs to be provided")
                elif ioe.errno == 404:
                    abort(f"{username} not found")
                else:
                    abort(f"Unexpected error: {ioe}")

    f.close()

if __name__ == "__main__":
    main()



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HSDS (HDF5 Scalable Data Service), Libraries and      #
# Utilities.  The full HSDS copyright notice, including                      #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################
import os
import sys
import json
import logging

default_cfg = {
    "hs_endpoint": {
        "default": None,
        "flags": ["-e", "--endpoint"],
        "help": "server endpoint, e.g. http://hsdshdflab.hdfgroup.org",
        "choices": ["ENDPOINT",]
    },
    "hs_username": {
        "default": None,
        "flags": ["-u", "--user"],
        "help": "user name credential",
        "choices": ["USERNAME",]
    },
    "hs_password": {
        "default": None,
        "flags": ["-p", "--password"],
        "help": "password credential",
        "choices": ["PASSWORD",]
    },
    "hs_api_key": {
        "default": None,
        "flags": ["--api_key",],
        "help": "user api key",
        "choices": ["API_KEY"]
    }, 
    "hs_bucket": {
        "default": None,
        "flags": ["--bucket",],
        "help": "storage Bucket to use (S3 Bucket, Azure Container, or top-level directory)",
        "choices": ["BUCKET",]
    },

    "loglevel": {
        "default": "error",
        "flags": ["--loglevel",],
        "help": "logging verbosity",
        "choices": ["debug", "info", "warning", "error"],
    },
    "logfile": {
        "default": None,
        "flags": ["--logfile",],
        "help": "file to send logout to (otherwise stdout)",
        "choices": ["FILENAME",]
    },
    "verbose": {
        "default": False,
        "flags": ["--verbose", "-v"],
        "help": "verbose output",
    },
    "ignore": {
        "default": False,
        "flags": ["--ignore",],
        "help": "don't exit on error"
    } 
}

hscmds = ("hsinfo", "hsconfigure", "hsls", "hstouch", "hsload", "hsget", "hsacl", "hsrm", "hsdiff")

class Config:
    """
    User Config state
    """
    def __init__(self, config_file=None, custom_entries=[], **kwargs):
        self._names = []
        self._values = {}
        self._flags = {}
        self._help = {}
        self._choices = {}
        self._flag_map = {}

        # set default entries
        for defaults in (default_cfg, custom_entries):
            for name in defaults:
                if name in self._names:
                    raise ValueError(f"config {name} already set")
                entry = defaults[name]
                self._names.append(name)
                if "default" in entry:
                    self._values[name] = entry["default"]
                if "flags" in entry:
                    self._flags[name] = entry["flags"]
                    for flag in entry["flags"]:
                        self._flag_map[flag] = name
                if "help" in entry:
                    self._help[name] = entry["help"]
                if "choices" in entry:
                    self._choices[name] = entry["choices"]

        if config_file:
            self._config_file = config_file
        elif os.path.isfile(".hscfg"):
            self._config_file = ".hscfg"
        else:
            self._config_file = os.path.expanduser("~/.hscfg")
        # process config file if found
        if os.path.isfile(self._config_file):
            line_number = 0
            with open(self._config_file) as f:
                for line in f:
                    line_number += 1
                    s = line.strip()
                    if not s:
                        continue
                    if s[0] == '#':
                        # comment line
                        continue
                    fields = s.split('=')
                    if len(fields) < 2:
                        print(f"config file: {self._config_file} line: {line_number} is not valid")
                        continue
                    k = fields[0].strip()
                    v = fields[1].strip()
                    if k not in self._names:
                        raise ValueError(f"undefined option: {name}")
                    if k in self._choices:
                        choices = self._choices[k]
                        if len(choices) > 1 and v not in self._choices:
                            raise ValueError(f"option {k} must be one of {choices}")
                    self._values[k] = v
        # override any config values with environment variable if found
        for k in self._names:
            if k.upper() in os.environ:
                v = os.environ[k.upper()]
                if name in self._choices:
                    choices = self._choices[name]
                    if len(choices) > 1 and v not in self._choices:
                        raise ValueError(f"option {name} must be one of {choices}")
                self._values[k] = v

        # finally update any values that are passed in to the constructor
        for name in kwargs.keys():
            if name in self._names:
                v = kwargs[name]
                if name in self._choices:
                    choices = self._choices[name]
                    if len(choices) > 1 and v not in self._choices:
                        raise ValueError(f"option {name} must be one of {choices}")
                self._values[name] = kwargs[name]

    def __getitem__(self, name):
        """ Get a config item  """
        if name not in self._names:
            return None
        return self._values[name]

    def setitem(self, name, value, flags=None, choices=None, help=None):
        """ Set a config item """
        if name not in self._names:
            self._names.append(name)
        self._values[name] = value 
        if flags is not None:
            self._flags[name] = flags
            for flag in flags:
                self._flag_map[flag] = name
        if choices is not None:
            self._choices[name] = choices
        if help is not None:
            self._help[name] = help

    def __setitem__(self, name, value):
        self.setitem(name, value)
     
    def __len__(self):
        return len(self._names)

    def __iter__(self):
        """ Iterate over config names """
        for name in self._names:
            yield name

    def __contains__(self, name):
        return name in self._names

    def __repr__(self):
        return json.dumps(self._values)

    def keys(self):
        return self._names

    def get_flags(self, name):
        if name in self._flags:
            return self._flags[name]
        else:
            return None

    def get_help(self, name):
        if name in self._help:
            return self._help[name]
        else:
            return None

    def get_see_also(self, this_cmd):
        msg = "See also the commands: "
        for cmd in hscmds:
            if cmd != this_cmd:
                msg += f"{cmd}, "
        msg = msg[:-2]  # remove trailing comma
        return msg
        

    def get_help_message(self, name):
        help_text= self.get_help(name)
        flags = self.get_flags(name)
        choices = self.get_choices(name)
        if not help_text or len(flags) == 0:
            return None
        
        msg = flags[0]
        for i in range(1, len(flags)):
            msg += f", {flags[i]}"
        if choices:
            if len(choices) == 1:
                msg += f" {choices[0]}"
            else:
                msg += " {" 
                for choice in choices:
                    msg += f"{choice}|"
                msg = msg[:-1]
                msg += "}"
        if len(msg) < 40:
            pad = " "*(40 - len(msg))
            msg += pad
        
        msg += f" {help_text}"
        
        return msg
        

    def get_nargs(self, name):
        choices = self._choices.get(name)
        if not choices:
            return 0
        else:
            return 1

    def get_choices(self, name):
        if name in self._choices:
            return self._choices[name]
        else:
            return 0

    def get_names(self):
        return self._names

    def set_cmd_flags(self, args, allow_post_flags=False):
        """ process any command line options
            return any place argument as a list 
        """
        options = []
        argn = 0
        while argn < len(args):
            arg = args[argn]
            val = None
            if len(args) > argn + 1:
                val = args[argn+1]
            if not arg.startswith("-"):
                options.append(arg)
                argn += 1
            elif options:
                if allow_post_flags:
                    options.append(arg)
                    argn += 1
                else:
                    raise ValueError("flags must be set before positional arguments")
            else:
                name = self._flag_map.get(arg)
                if arg in ("-h", "--help"):
                    raise ValueError()  # trigger print usage
                if name not in self._names:
                    raise ValueError("option not found")
                if not self.get_nargs(name):
                    # set flag
                    self._values[name] = True
                    argn += 1
                else:
                    if not val:
                        raise ValueError("option value missing")
                    if self._choices.get(name):
                        choices = self._choices.get(name)

                        if choices and len(choices) > 1 and val not in self._choices.get(name):
                            raise ValueError(f"option value must be one of {self._choices.get(name)}")
                    self._values[name] = val
                    argn += 2
        return options
    
    def get_loglevel(self):
        val = self._values["loglevel"]
        val = val.upper()
        choices = ("DEBUG", "INFO", "WARNING", "ERROR")
        if val == "DEBUG":
            loglevel = logging.DEBUG
        elif val == "INFO":
            loglevel = logging.INFO
        elif val in ("WARN", "WARNING"):
            loglevel = logging.WARNING
        elif val == "ERROR":
            loglevel = logging.ERROR
        else:
            raise ValueError(f"loglevel must be one of {choices}")
        return loglevel

    def get_cmd(self):
        """ return command argument used to invoke"""
        cmd = sys.argv[0].split('/')[-1]
        if cmd.endswith(".py"):
            cmd = "python " + cmd
        return cmd

    def print(self, msg):
        if self._values.get("logfile"):
            # write msg to logfile as info
            logging.info(msg)
        if self._values.get("verbose"):
            print(msg)
         
    











import json

# rom .config import Config

"""
get aiobotocore lambda client
"""

LAMBDA_REQ_PREFIX = "http+lambda://"

STATUS_REASONS = {
    200: "OK",
    201: "Created",
    202: "Accepted",
    204: "No Content",
    400: "Bad Request",
    401: "Unauthorized",
    403: "Forbidden",
    404: "Not Found",
    408: "Request Timeout",
    409: "Confict",
    410: "Gone",
    413: "Payload Too Large",
    500: "Internal Server Error",
    501: "Not Implemented",
    503: "Service Unavailable",
    504: "Gateway Timeout",
    507: "Insufficient Storage",
}


class HttpChunkIterator(object):
    """
    Class to iterate through list of chunks of a http response
    """

    def __init__(self, data, chunk_size=1):
        self._data = data
        self._chunk_size = chunk_size
        self._index = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self._data is None:
            raise StopIteration()
        if self._index >= len(self._data):
            raise StopIteration()
        num_bytes = len(self._data)
        n = self._index
        m = n + self._chunk_size
        if m > num_bytes:
            m = num_bytes
        self._index = m
        return self._data[n:m]


class LambdaResponse:
    def __init__(self, lambda_rsp):
        self._status_code = 500
        self._reason = ""
        self._headers = {}
        self._text = None
        self._json = None
        self._content_length = 0
        self._iter_index = 0
        if lambda_rsp and isinstance(lambda_rsp, dict):
            if "StatusCode" in lambda_rsp:
                lambda_status_code = lambda_rsp["StatusCode"]

                if lambda_status_code in (200, 201) and "Payload" in lambda_rsp:
                    payload = lambda_rsp["Payload"]
                    rsp_text = payload.read().decode("utf-8")
                    rsp_payload = json.loads(rsp_text)
                    if rsp_payload.get("isBase64Encoded"):
                        is_base64_encoded = True
                    else:
                        is_base64_encoded = False

                    if "statusCode" in rsp_payload:
                        self._status_code = rsp_payload["statusCode"]
                    if "headers" in rsp_payload:
                        headers_text = rsp_payload["headers"]

                        headers = json.loads(headers_text)
                        for k in headers:
                            v = headers[k]
                            self._headers[k] = v
                    if self._status_code in (200, 201) and "body" in rsp_payload:
                        body_text = rsp_payload["body"]
                        # set the json prop for a dict,
                        # otherwise just set the text prop
                        if isinstance(body_text, dict):
                            self._json = body_text
                        elif is_base64_encoded and body_text:
                            # convert hex encoded to bytes
                            self._text = bytes.fromhex(body_text)
                        else:
                            self._text = body_text

                else:
                    raise ValueError("lambda: unable to get payload")
        else:
            raise TypeError("lambda: expected dict response")
        if self._status_code in STATUS_REASONS:
            self._reason = STATUS_REASONS[self._status_code]
        else:
            self._reason = "Unexpected status code"

    @property
    def status_code(self):
        return self._status_code

    @property
    def reason(self):
        return self._reason

    @property
    def text(self):
        if self._text:
            return self._text
        elif self._json:
            self._text = json.dumps(self._json)
            return self._text
        else:
            return None

    def json(self):
        if self._json:
            return self._json
        elif self._text:
            self._json = json.loads(self._text)
            return self._json
        else:
            return None

    def iter_content(self, chunk_size=1):
        return HttpChunkIterator(self._text, chunk_size=chunk_size)

    @property
    def headers(self):
        return self._headers

    @property
    def content_length(self):
        if self._text:
            return len(self._text)
        elif self._json:
            self._text = json.dumps(self._text)
            return len(self._text)
        else:
            return 0


class Session:
    def __init__(self, timeout=10):
        self.timeout = timeout

    def __enter__(self):
        pass

    def __exit__(self):
        pass

    def mount(self, protocol, adapter):
        # TBD
        # print(f"requests_lambda mount({protocol})")
        pass

    def _invoke(self, req, method="GET", params=None, headers=None, data=None):
        if not req:
            msg = "no req"
            raise ValueError(msg)
        if not req.startswith(LAMBDA_REQ_PREFIX):
            msg = f"Expected req to start with {LAMBDA_REQ_PREFIX}"
            raise ValueError(msg)
        if method not in ("GET", "PUT", "POST", "DELETE"):
            msg = f"Unexpected method: {method}"
            raise ValueError(msg)
        if method in ("GET", "DELETE") and data:
            msg = f"data not expected for method: {method}"
            raise ValueError(msg)

        # Convert uri of the form: http+lambda://FUNC_NAME/REQ
        # as:
        # function_name = FUNC_NAME
        # req_path = REQ
        # params = {PARAMS}
        s = req[len(LAMBDA_REQ_PREFIX) :]  # strip off protocol
        index = s.find("/")
        if index <= 0:
            msg = "Unexpected request"
            raise ValueError(msg)
        function_name = s[:index]
        if function_name.find("/") >= 0:
            msg = f"unexpected lambda function name: {function_name}"
            raise ValueError(msg)
        index = s.find(function_name)
        req_path = s[index + len(function_name) :]
        if not req_path:
            msg = "no request path found"
            raise ValueError(msg)

        # convert header values to string from bytes if needed
        json_headers = {}
        for k in headers:
            v = headers[k]
            if isinstance(v, bytes):
                json_headers[k] = v.decode("utf-8")
            else:
                json_headers[k] = v

        req_json = {
            "method": method,
            "path": req_path,
            "params": params,
            "headers": json_headers,
            "body": data,
        }

        payload = json.dumps(req_json).encode("utf-8")

        import boto3  # import here so it's not a global dependency
        from botocore.exceptions import ClientError

        # with boto3.client('lambda')
        lambda_client = boto3.client("lambda")
        try:
            lambda_rsp = lambda_client.invoke(
                FunctionName=function_name,
                InvocationType="RequestResponse",
                Payload=payload,
            )
        except ClientError as ce:
            if "Error" in ce.response and "Code" in ce.response["Error"]:
                error_code = ce.response["Error"]["Code"]
            else:
                error_code = "Unknown Lambda error"
            if error_code == "UnrecognizedClientException":
                # this happens when the AWS access key not provided
                error_code += " (are the AWS credentials valid?)"
            raise ValueError(error_code)
        rsp = LambdaResponse(lambda_rsp)
        return rsp

    def get(
        self, req, params=None, headers=None, stream=False, timeout=None, verify=None
    ):
        """
        Lambda GET request

        req should be in form: "http+lambda://function/path"
        """
        if stream:
            raise ValueError("stream not supported for Lambda")
        rsp = self._invoke(req, params=params, headers=headers)
        return rsp

    def put(self, req, params=None, headers=None, data=None, verify=None):
        """
        Lambda PUT request

        req should be in form: "http+lambda://function/path"
        """
        rsp = self._invoke(req, method="PUT", params=params, headers=headers, data=data)
        return rsp

    def post(self, req, params=None, headers=None, data=None, verify=None):
        """
        Lambda POST request

        req should be in form: "http+lambda://function/path"
        """
        rsp = self._invoke(
            req, method="POST", params=params, headers=headers, data=data
        )
        return rsp

    def delete(self, req, params=None, headers=None, verify=None):
        """
        Lambda DELETE request

        req should be in form: "http+lambda://function/path"
        """
        rsp = self._invoke(req, method="DELETE", params=params, headers=headers)
        return rsp

    def close(self):
        # TBD - release any held resources
        pass



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HSDS (HDF5 Scalable Data Service), Libraries and      #
# Utilities.  The full HSDS copyright notice, including                      #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

import sys
import logging

try:
    import h5py
    import h5pyd
    import numpy as np
except ImportError as e:
    sys.stderr.write(f"ERROR : {e} : install it to use this utility...")
    sys.exit(1)

# copy rather than link for any datasets with product of extents less than the following
MIN_DSET_ELEMENTS_FOR_LINKING=512

# adjust chunk shape to fit between min and max chunk sizes when possible
MIN_CHUNK_SIZE = 1 * 1024 * 1024
MAC_CHUNK_SIZE = 8 * 1024 * 1024

H5Z_FILTER_MAP = { 32001: "blosclz", 
                   32004: "lz4",
                   32008: "bitshuffle", 
                   32015: "zstd",
}


# check if hdf5 library version supports chunk iteration
hdf_library_version  = h5py.version.hdf5_version_tuple
library_has_chunk_iter = (hdf_library_version >= (1, 14, 0) or (hdf_library_version < (1, 12, 0) and (hdf_library_version >= (1, 10, 10))))

def dump_dtype(dt):
    if not isinstance(dt, np.dtype):
        raise TypeError(f"expected np.dtype, but got: {type(dt)}")
    if len(dt) > 0:
        out = "{"
        for name in dt.fields:
            subdt = dt.fields[name][0]
            out += f"{name}: {dump_dtype(subdt)} |"
        out = out[:-1] + "}"
    else:
        ref = h5py.check_dtype(ref=dt)
        if ref:
            out = str(ref)
        else:
            vlen = h5py.check_dtype(vlen=dt)
            if vlen:
                out = "VLEN: " + dump_dtype(vlen)
            else:
                out = str(dt)
    return out


def is_h5py(obj):
    # Return True if objref is a h5py object and False is not
    if isinstance(obj, object) and isinstance(obj.id.id, int):
        return True
    else:
        return False


def is_reference(val):
    try:
        if isinstance(val, object) and val.__class__.__name__ == "Reference":
            return True
        elif isinstance(val, type) and val.__name__ == "Reference":
            return True
    except AttributeError as ae:
        msg = f"is_reference for {val} error: {ae}"
        logging.warning(msg)

    return False


def is_regionreference(val):
    try:
        if isinstance(val, object) and val.__class__.__name__ == "RegionReference":
            return True
        elif isinstance(val, type) and val.__name__ == "RegionReference":
            return True
    except AttributeError as ae:
        msg = f"is_regionreference for {val} error: {ae}"
        logging.warning(msg)

    return False


def has_reference(dtype):
    has_ref = False
    if not isinstance(dtype, np.dtype):
        return False
    if len(dtype) > 0:
        for name in dtype.fields:
            item = dtype.fields[name]
            if has_reference(item[0]):
                has_ref = True
                break
    elif dtype.metadata and "ref" in dtype.metadata:
        basedt = dtype.metadata["ref"]
        has_ref = is_reference(basedt)
    elif dtype.metadata and "vlen" in dtype.metadata:
        basedt = dtype.metadata["vlen"]
        has_ref = has_reference(basedt)
    return has_ref

def get_reftype(obj):
    if is_h5py(obj):
        ref_type = h5py. special_dtype(ref=h5py.Reference)
    else:
        ref_type = h5pyd.special_dtype(ref=h5pyd.Reference)
    return ref_type


def is_vlen(dtype):
    if dtype.metadata and "vlen" in dtype.metadata:
        return True
    else:
        return False


def get_fillvalue(dset):

    fillvalue = None
    if not is_vlen(dset.dtype):
        try:
            # can trigger a runtime error if fillvalue is undefined
            fillvalue = dset.fillvalue
        except RuntimeError:
            logging.warning(f"runtime error getting fillvalue for dataset: {dset.name}")
    return fillvalue


def is_compact(dset):
    if isinstance(dset.id.id, str):
        return False  # compact storage not used with HSDS
    cpl = dset.id.get_create_plist()
    if cpl.get_layout() == h5py.h5d.COMPACT:
        return True
    else:
        return False


def convert_dtype(srcdt, ctx):
    """Return a dtype based on input dtype, converting any Reference types from
    h5py style to h5pyd and vice-versa.
    """

    msg = f"convert dtype: {srcdt}, type: {type(srcdt)}"
    logging.info(msg)

    if len(srcdt) > 0:
        fields = []
        for name in srcdt.fields:
            item = srcdt.fields[name]
            # item is a tuple of dtype and integer offset
            field_dt = convert_dtype(item[0], ctx)
            fields.append((name, field_dt))
        tgt_dt = np.dtype(fields)
    else:
        # check if this a "special dtype"
        if srcdt.metadata and "ref" in srcdt.metadata:
            ref = srcdt.metadata["ref"]
            if is_reference(ref):
                if is_h5py(ctx["fout"]):
                    tgt_dt = h5py.special_dtype(ref=h5py.Reference)
                else:
                    tgt_dt = h5pyd.special_dtype(ref=h5pyd.Reference)
            elif is_regionreference(ref):
                if is_h5py(ctx["fout"]):
                    tgt_dt = h5py.special_dtype(ref=h5py.RegionReference)
                else:
                    tgt_dt = h5py.special_dtype(ref=h5py.RegionReference)
            else:
                msg = f"Unexpected ref type: {srcdt}"
                logging.error(msg)
                if not ctx["ignore_error"]:
                    raise TypeError(msg)
        elif srcdt.metadata and "vlen" in srcdt.metadata:
            src_vlen = srcdt.metadata["vlen"]
            if isinstance(src_vlen, np.dtype):
                tgt_base = convert_dtype(src_vlen, ctx)
            else:
                tgt_base = src_vlen
            if is_h5py(ctx["fout"]):
                tgt_dt = h5py.special_dtype(vlen=tgt_base)
            else:
                tgt_dt = h5pyd.special_dtype(vlen=tgt_base)
        elif srcdt.kind == "U":
            # use vlen for unicode strings
            if is_h5py(ctx["fout"]):
                tgt_dt = h5py.special_dtype(vlen=str)
            else:
                tgt_dt = h5pyd.special_dtype(vlen=str)
        else:
            tgt_dt = srcdt
    return tgt_dt


# ----------------------------------------------------------------------------------
def copy_element(val, src_dt, tgt_dt, ctx):
    msg = f"copy_element, val: {val} "
    msg += f"val type: {type(val)} src_dt:  {dump_dtype(src_dt)} "
    msg += f" tgt_dt: {dump_dtype(tgt_dt)}"
    logging.debug(msg)

    fin = ctx["fin"]
    fout = ctx["fout"]
    out = None
    if len(src_dt) > 0:
        out_fields = []
        i = 0
        for name in src_dt.fields:
            field_src_dt = src_dt.fields[name][0]
            field_tgt_dt = tgt_dt.fields[name][0]
            field_val = val[i]
            i += 1
            out_field = copy_element(field_val, field_src_dt, field_tgt_dt, ctx)
            out_fields.append(out_field)
            out = tuple(out_fields)
    elif src_dt.metadata and "ref" in src_dt.metadata:
        if not tgt_dt.metadata or "ref" not in tgt_dt.metadata:
            raise TypeError(f"Expected tgt dtype to be ref, but got: {tgt_dt}")
        ref = tgt_dt.metadata["ref"]
        if is_reference(ref):
            # initialize out to null ref
            if is_h5py(ctx["fout"]):
                out = h5py.Reference()  # null h5py ref
            else:
                out = ""  # h5pyd refs are strings

            if ref and val: 
                try:
                    fin_obj = fin[val]
                except AttributeError as ae:
                    msg = f"Unable able to get obj for ref value: {ae}"
                    logging.error(msg)
                    if not ctx["ignore_error"]:
                        raise IOError(msg)
                    return None

                h5path = fin_obj.name
                if not h5path:
                    msg = "No path found for ref object"
                    logging.warn(msg)
                    if ctx["verbose"]:
                        print(msg)
                else:
                    fout_obj = fout[h5path]
                    if is_h5py(ctx["fout"]):
                        out = fout_obj.ref
                    else:
                        # convert to string for JSON serialization
                        out = str(fout_obj.ref)


        elif is_regionreference(ref):
            out = "tbd"
        else:
            raise TypeError(f"Unexpected ref type: {type(ref)}")
    elif src_dt.metadata and "vlen" in src_dt.metadata:
        logging.debug(
            "copy_elment, got vlen element, dt: {}".format(src_dt.metadata["vlen"])
        )
        if not isinstance(val, np.ndarray):
            raise TypeError(f"Expecting ndarray or vlen element, but got: {type(val)}")
        if not tgt_dt.metadata or "vlen" not in tgt_dt.metadata:
            raise TypeError(f"Expected tgt dtype to be vlen, but got: {tgt_dt}")
        src_vlen_dt = src_dt.metadata["vlen"]
        tgt_vlen_dt = tgt_dt.metadata["vlen"]
        if has_reference(src_vlen_dt):
            if len(val.shape) == 0:
                # scalar array
                e = val[()]
                v = copy_element(e, src_vlen_dt, tgt_vlen_dt, ctx)
                out = np.array(v, dtype=tgt_dt)
            else:

                out = np.zeros(val.shape, dtype=tgt_dt)
                for i in range(len(out)):
                    e = val[i]
                    out[i] = copy_element(e, src_vlen_dt, tgt_vlen_dt, ctx)
        else:
            # can just directly copy the array
            out = np.zeros(val.shape, dtype=tgt_dt)
            out[...] = val[...]
    else:
        out = val  # can just copy as is
    return out


# ---------------------------------------------------------------------------------
def copy_array(src_arr, ctx):
    """Copy the numpy array to a new array.
    Convert any reference type to point to item in the target's hierarchy.
    """
    if not isinstance(src_arr, np.ndarray):
        raise TypeError(f"Expecting ndarray, but got: {src_arr}")
    tgt_dt = convert_dtype(src_arr.dtype, ctx)
    tgt_arr = np.zeros(src_arr.shape, dtype=tgt_dt)

    if has_reference(src_arr.dtype):
        # flatten array to simplify iteration
        count = np.product(src_arr.shape)
        tgt_arr_flat = tgt_arr.reshape((count,))
        src_arr_flat = src_arr.reshape((count,))
        for i in range(count):
            e = src_arr_flat[i]
            element = copy_element(e, src_arr.dtype, tgt_dt, ctx)
            tgt_arr_flat[i] = element
        tgt_arr = tgt_arr_flat.reshape(src_arr.shape)
    else:
        # can just copy the entire array
        tgt_arr[...] = src_arr[...]
    return tgt_arr


# ----------------------------------------------------------------------------------
def copy_attribute(desobj, name, srcobj, ctx):

    msg = f"creating attribute {name} in {srcobj.name}"
    logging.info(msg)

    if ctx["verbose"]:
        print(msg)

    tgtarr = None
    data = srcobj.attrs[name]
    src_dt = None
    
    # check for non-numpy types that might get returned
    if is_regionreference(data):
        msg = "regionreference types not supported, "
        msg += f"attribute {name} in object {desobj.name} will not be loaded"
        if ctx["verbose"]:
            print(msg)
        logging.warning(msg)
        return

    if is_reference(data):
        src_dt = get_reftype(srcobj)
        tgt_dt = get_reftype(desobj)
        tgt_ref = copy_element(data, src_dt, tgt_dt, ctx)
        try:
            desobj.attrs.create(name, tgt_ref)
        except (IOError, TypeError) as e:
            msg = f"ERROR: failed to create attribute {name} "
            msg += f"of object {desobj.name} for reference type -- {e}"
            logging.error(msg)
            if not ctx["ignore_error"]:
                raise IOError(msg)

        # done with non-numpy compatible data
        return

    try:
        src_dt = data.dtype
    except AttributeError:
        # convert to numpy type
        data = np.asarray(data)
        src_dt = data.dtype
        
    if src_dt.kind == "S" and isinstance(data, bytes):
        # check that this is actually utf-encodable
        try:
            data.decode("utf-8")
        except UnicodeDecodeError:
            msg = f"byte value for attribute {name} in {srcobj.name} "
            msg += "is not utf8 encodable - using surrogateescaping"
            logging.warning(msg)
            if ctx["verbose"]:
                print(msg)
            data = data.decode("utf-8", errors="surrogateescape")
            src_dt = None  # let numpy figure out the unicode type

    # First, make sure we have a NumPy array.
    if is_h5py(srcobj):
        src_empty = h5py.Empty
    else:
        src_empty = h5pyd.Empty
    if is_h5py(desobj):
        des_empty = h5py.Empty
    else:
        des_empty = h5pyd.Empty

    
    if isinstance(data, src_empty):
        # create Empty object with tgt dtype
        tgt_dt = convert_dtype(src_dt, ctx)
        tgtarr = des_empty(tgt_dt)
    else:
        srcarr = np.asarray(data, order="C", dtype=src_dt)
        tgtarr = copy_array(srcarr, ctx)

    try:
        desobj.attrs.create(name, tgtarr)
    except (IOError, TypeError) as e:
        msg = f"ERROR: failed to create attribute {name} "
        msg += f"of object {desobj.name} -- {e}"
        logging.error(msg)
        if not ctx["ignore_error"]:
            raise IOError(msg)

# "safe" resize method where new extent can be <= existing extent
def resize_dataset(dset, extent, axis=0):
    logging.debug(f"resize_dataset {dset} to {extent}")
    try:
        dset.resize(extent, axis=axis)
    except IOError:
        # raise this if it's not a case where the extent is already increased
        if dset.shape[axis] < extent:
            raise

# ----------------------------------------------------------------------------------
def get_chunk_layout(dset):
    if is_h5py(dset):
        msg = "get_chunk_layout called on hdf5 dataset"
        logging.error(msg)
        raise IOError(msg)
    dset_json = dset.id.dcpl_json
    if "layout" not in dset_json:
        msg = f"expect to find layout key in dset_json: {dset_json}"
        logging.error(msg)
        raise IOError(msg)
    layout = dset_json["layout"]
    logging.debug(f"got chunk layout for dset id: {dset.id.id}: {layout}")
    return layout
    


# ----------------------------------------------------------------------------------
def get_chunk_dims(dset):
    if dset.chunks:
        chunk_dims = dset.chunks
    elif not is_h5py(dset):
        # h5pyd datset, check layout
        layout = dset.id.layout
        if layout and "dims" in layout:
            # linked datasets will use chunk shape in source hdf5 file
            # (or source dataset shape for contiguous datasets)
            chunk_dims = tuple(layout["dims"])
        else:
            # define a psuedo-chunk with same dimensions as the dataset
            chunk_dims = dset.shape
    elif dset.shape is None:
        chunk_dims = None
    elif len(dset.shape) > 0 and np.prod(dset.shape) > 0:
        # just use dataset shape
        chunk_dims = dset.shape
    else:
        # no chunk dims
        chunk_dims = None
    return chunk_dims

# ----------------------------------------------------------------------------------
def get_chunktable_dims(dset):
    rank = len(dset.shape)
    chunk_dims = get_chunk_dims(dset)
    table_dims = []
    for dim in range(rank):
        dset_extent = dset.shape[dim]
        chunk_extent = chunk_dims[dim]

        if dset_extent > 0 and chunk_extent > 0:
            table_extent = -(dset_extent // -chunk_extent)
        else:
            table_extent = 0
        table_dims.append(table_extent)
    table_dims = tuple(table_dims)
    return table_dims

# ----------------------------------------------------------------------------------
def get_num_chunks(dset):
    if dset.shape is None:
        # null space dataset - no data
        return 0
    elif np.prod(dset.shape) == 0:
        # zero extent dataset
        return 0
    elif np.prod(dset.shape) == 1:
        # scalar or extent 1 dataset
        return 1
    elif dset.chunks:
        if is_h5py(dset):
            dsetid = dset.id
            spaceid = dsetid.get_space()
            try:
                num_chunks = dsetid.get_num_chunks(spaceid)
            except ValueError:
                # thrown for compact or contiguous datasets,
                # just treat as 1 chunk
                return 1
        else:
            # for hsds, just return maximum number of chunk in dataset
            chunk_table_dims = get_chunktable_dims(dset)
            num_chunks = np.prod(chunk_table_dims)
    else:
        # use a psuedo-chunk with same dimensions as the dataset
        num_chunks = 1
    return num_chunks

# ----------------------------------------------------------------------------------
def get_dset_offset(dset):
    """ Return dataset file offset for HDF5 contiguous datasets """
    if not is_h5py(dset):
        return -1
    if dset.chunks is not None:
        return -1
    offset = dset.id.get_offset()
    if offset is None:
        return -1
    return offset
   
# ----------------------------------------------------------------------------------
def get_chunktable_dtype(include_file_uri=False):
    if include_file_uri:
        dt_str = h5pyd.special_dtype(vlen=bytes)
        dt = np.dtype([("offset", np.int64), ("size", np.int32), ("file_uri", dt_str)])
    else:
        dt = np.dtype([("offset", np.int64), ("size", np.int32)])
    return dt

# ----------------------------------------------------------------------------------
def get_chunk_table_index(chunk_offset, chunk_dims):
    if len(chunk_offset) != len(chunk_dims):
        msg = f"Unexptected chunk offset: {chunk_offset}"
        logging.error(msg)
    rank = len(chunk_offset)
    chunk_index = []
    for i in range(rank):
        chunk_index.append(chunk_offset[i]//chunk_dims[i])
    return tuple(chunk_index)

# ----------------------------------------------------------------------------------
def get_chunk_locations(dset, ctx, include_file_uri=False):
    if not is_h5py(dset):
        msg = "get_chunklocations should only be used with HDF5 datasets"
        logging.error(msg)
        if ctx["ignore_error"]:
            return None
        else:
            raise IOError(msg)
    
    if dset.chunks is None:
        msg = "get_chunklocations - dataset is not chunked"
        logging.error(msg)
        if ctx["ignore_error"]:
            return None
        else:
            raise IOError(msg)
    
    rank = len(dset.shape)

    spaceid = dset.id.get_space()
    logging.debug(f"using chunk_iter: {library_has_chunk_iter}")

    dt = get_chunktable_dtype(include_file_uri=include_file_uri)

    chunktable_dims = get_chunktable_dims(dset)
    logging.debug(f"chunktable_dims: {chunktable_dims}")

    if chunktable_dims is None:
        msg = "no chunktable dimension returned"
        logging.error(msg)
        if ctx["ignore_error"]:
            return None
        else:
            raise IOError(msg)

    chunk_arr = np.zeros(chunktable_dims, dtype=dt)

    if include_file_uri:
        s3path = ctx["s3path"].encode("utf-8")
    else:
        s3path = None

    if dset.chunks is None:
        chunk_offset = get_dset_offset(dset)
        if np.prod(chunktable_dims) != 1:
            msg = "Expected one chunk"
            logging.error(msg)
            if not ctx["ignore_error"]:
                raise IOError(msg)
        if chunk_offset <= 0:
            msg = "Expected dset_offset to be greater than zero"
            logging.error(msg)
            if not ctx["ignore_error"]:
                raise IOError(msg)

        chunk_size = dset.id.get_storage_size()
        if include_file_uri:
            e = (chunk_offset, chunk_size, s3path)
        else:
            e = (chunk_offset, chunk_size)
    
        chunk_arr[...] = e

        # return one-element array
        return chunk_arr
    
    # get chunk locations for non-contiguous datasets
    chunk_dims = get_chunk_dims(dset)

    if library_has_chunk_iter:
        def init_chunktable_callback(chunk_info):
            # Use chunk offset as index 
            index = get_chunk_table_index(chunk_info[0], chunk_dims)
            byte_offset = chunk_info[2]
            chunk_size = chunk_info[3]

            if not isinstance(index, tuple) or len(index) != rank:
                msg = f"Unexpected array_offset: {index} for dataset with rank: {rank}"
                logging.error(msg)
                raise IOError(msg)
            
            if include_file_uri:
                e = (byte_offset, chunk_size, s3path)
            else:
                e = (byte_offset, chunk_size)

            chunk_arr[index] = e
                
        dset.id.chunk_iter(init_chunktable_callback)
    else: 
        # Using old HDF5 version without H5Dchunk_iter
        num_chunks = get_num_chunks(dset)


        for i in range(num_chunks):
            chunk_info = dset.id.get_chunk_info(i, spaceid)
            index = get_chunk_table_index(chunk_info[0], chunk_dims)
            byte_offset = chunk_info[2]
            chunk_size = chunk_info[3]
            logging.debug(f"got chunk_info: {chunk_info} for chunk: {i}")
            if not isinstance(index, tuple) or len(index) != rank:
                msg = f"Unexpected array_offset: {index} for dataset with rank: {rank}"
                logging.error(msg)
                raise IOError(msg)
            
            if include_file_uri:
                e = (byte_offset, chunk_size, s3path)
            else:
                e = (byte_offset, chunk_size)
            
            chunk_arr[index] = e
            
            if i % 5000 == 0:
                logging.info(f"{i} chunks indexed")

    return chunk_arr

# ----------------------------------------------------------------------------------
def create_chunktable(dset, dset_dims, ctx):
    logging.debug(f"create_chunktable({dset}, {dset_dims}")

    num_chunks = get_num_chunks(dset)
    rank = len(dset_dims)
    extend = True if rank > len(dset.shape) else False
    if extend:
        include_file_uri = True
    else:
        include_file_uri = False

    logging.debug(f"num_chunks: {num_chunks}")

    chunks = {}  # pass a map to create_dataset

    if num_chunks > 10 or extend:
        # create anonymous dataset to hold chunk info
        dt = get_chunktable_dtype(include_file_uri=include_file_uri)

        chunktable_dims = [0,] if extend else []
        
        chunktable_dims.extend(get_chunktable_dims(dset))
        chunktable_dims = tuple(chunktable_dims)
        logging.debug(f"chunktable_dims: {chunktable_dims}")
        chunktable_maxshape = [None,] if extend else []
        chunktable_maxshape.extend(get_chunktable_dims(dset))
        chunk_dims = [1,] if extend else []
        chunk_dims.extend(get_chunk_dims(dset))

        fout = ctx["fout"]
        kwargs = {}
        kwargs["shape"] = chunktable_dims
        kwargs["dtype"] = dt
        kwargs["maxshape"] = chunktable_maxshape
        if ctx["dataload"] == "fastlink" and dset.name:
            kwargs["initializer"] = "chunklocator"
            initializer_opts = []
            initializer_opts.append(f"--h5path={dset.name}")
            linkpath = ctx["s3path"]
            s3prefix = "s3://"
            if linkpath.startswith(s3prefix):
                linkpath = linkpath[len(s3prefix):]
            index = linkpath.find("/")
            if index < 1:
                msg = f"unexpected linkpath: {linkpath}"
                logging.error(msg)
                raise ValueError(msg)
            bucket = linkpath[:index]
            filepath = linkpath[(index+1):]
            initializer_opts.append(f"--filepath={filepath}")
            initializer_opts.append(f"--bucket={bucket}")
            logging.info(f"using initializer: {initializer_opts}")
            kwargs["initializer_opts"] = initializer_opts
        
        anon_dset = fout.create_dataset(None, **kwargs)
        msg = f"created chunk table: {anon_dset}"
        logging.info(msg)
        if ctx["verbose"]:
            print(msg)
        chunks["class"] = "H5D_CHUNKED_REF_INDIRECT"
        if not extend:
            chunks["file_uri"] = ctx["s3path"]
        chunks["dims"] = chunk_dims
        chunks["chunk_table"] = anon_dset.id.id
        chunks["hyper_dims"] = dset.chunks

    elif num_chunks <= 1 and dset.chunks is None:
        # use contiguous mapping
        chunks["class"] = "H5D_CONTIGUOUS_REF"
        chunks["file_uri"] = ctx["s3path"]
        dset_offset = get_dset_offset(dset)
        if dset_offset <= 0:
            msg = f"unexpected dataset_offset: {dset_offset}"
            logging.error(msg)
            if not ctx["ignore_error"]:
                raise IOError(msg)
        dset_size = dset.id.get_storage_size()
        if dset_size <= 0:
            msg = f"unexpected dataset storage size: {dset_size}"
            logging.error(msg)
            if not ctx["ignore_error"]:
                raise IOError(msg)
        chunks["offset"] = dset_offset
        # TBD - check the size is not too large
        chunks["size"] = dset_size
    else:
        # construct map of chunks if count is less than 10
        # get the chunk locations
        chunk_map = {}
        chunk_dims = get_chunk_dims(dset)
        spaceid = dset.id.get_space()

        for i in range(num_chunks):
            chunk_info = dset.id.get_chunk_info(i, spaceid)
            index = chunk_info.chunk_offset
            logging.debug(f"got chunk_info: {chunk_info} for chunk: {i}")
            if not isinstance(index, tuple) or len(index) != rank:
                msg = f"Unexpected array_offset: {index} for dataset with rank: {rank}"
                logging.error(msg)
                if ctx["ignore_error"]:
                    continue
                else:
                    raise IOError(msg)
                
            chunk_key = ""
            for dim in range(rank):
                chunk_key += str(index[dim] // chunk_dims[dim])
                if dim < rank - 1:
                    chunk_key += "_"
            logging.debug(f"adding chunk_key: {chunk_key}")
            chunk_map[chunk_key] = (chunk_info.byte_offset, chunk_info.size)

        chunks["class"] = "H5D_CHUNKED_REF"
        if not extend:
            chunks["file_uri"] = ctx["s3path"]
        chunks["dims"] = dset.chunks
        chunks["chunks"] = chunk_map

    logging.info(f"using chunk layout: {chunks}")
    return chunks


# ----------------------------------------------------------------------------------
def update_chunktable(src, tgt, ctx):
    layout = tgt.id.layout
    if not layout:
        raise IOError("expected dataset layout to be set")
    if layout["class"] != "H5D_CHUNKED_REF_INDIRECT":
        logging.info("update_chunktable not supported for this chunk class")
        return
    if ctx["dataload"] == "fastlink":
        logging.info("skip update_chunktable for fastload")
        return
    rank = len(tgt.shape)
    chunktable_id = layout["chunk_table"]

    fout = ctx["fout"]
    logging.info(f"update_chunk_table {src.name}, id: {src.id.id}")

    # create a numpy array containing chunk refs for each chunk in src array
    extend = True if rank > len(src.shape) else False
    if extend and not ctx["s3path"]:
        msg = "expected s3path to be set for extend mode"
        logging.error(msg)
        if not ctx["ignore_error"]:
            raise IOError(msg)
        return

    chunktable = fout[f"datasets/{chunktable_id}"]
    chunk_arr = get_chunk_locations(src, ctx, include_file_uri=extend)

    msg = f"dataset chunk dimensions {chunktable.shape} not compatible with {chunk_arr.shape}"
    if len(chunktable.shape) == len(chunk_arr.shape):
        if chunktable.shape != chunk_arr.shape:
            logging.error(msg)
            if not ctx["ignore_error"]:
                raise IOError(msg)
            return
    elif len(chunk_arr.shape) + 1 == len(chunktable.shape):
        if chunk_arr.shape != chunktable.shape[1:]:
            logging.error(msg)
            return
    else:
        logging.error(msg)
        if not ctx["ignore_error"]:
            raise IOError(msg)
        return

    if not is_h5py(src):
        # hsds dataset
        if not extend:
            msg = "unexpected src type for update_chunktable"
            logging.error(msg)
            if not ctx["ignore_error"]:
                raise IOError(msg)
            return
    
        layout = get_chunk_layout(src)
        layout_class = layout["class"]
        if layout_class == "H5D_CONTIGUOUS_REF":
            chunk_offset = layout["offset"]
            chunk_size = layout["size"]
            file_uri = layout["file_uri"]
            chunk_arr = [(chunk_offset, chunk_size, file_uri),]
        elif layout_class == "H5D_CHUNKED_REF":
            file_uri = layout["file_uri"]
            chunkmap = layout["chunks"]  # e.g.{'0_2': [4016, 2000000]}}
            for k in chunkmap:
                v = chunkmap[k]
                v.append(file_uri)
                v = tuple(v)
                index = []
                chunk_indices = k.split("_")
                for i in range(len(chunk_indices)):
                    index.append(int(chunk_indices[i]))
                index = tuple(index)
                chunk_arr[index] = v
        elif layout_class == "H5D_CHUNKED_REF_INDIRECT":
            file_uri = layout["file_uri"] 
            orig_chunktable_id = layout["chunk_table"]
            orig_chunktable = fout[f"datasets/{orig_chunktable_id}"]
            # iterate through contents and add file uri
            arr = orig_chunktable[...]
            it = np.nditer(arr, flags=['multi_index'])
            for _ in it:
                value = arr[it.multi_index]
                if value[1] == 0:
                    # no chunk location set
                    continue
                e = list(value)
                e.append(file_uri)
                e = tuple(e)
                tgt_index = [0,]
                tgt_index.extend(it.multi_index)
                tgt_index = tuple(tgt_index)
                chunk_arr[it.multi_index] = e
        else:
            msg = f"expected chunk ref class but got: {layout_class}"
            logging.error(msg)
            if not ctx["ignore_error"]:
                raise IOError(msg)
            return
            
    if len(tgt.shape) > len(src.shape):
        # append mode, extend the first dimension of table by one
        extent = chunktable.shape[0] + 1
        resize_dataset(chunktable, extent)
        chunktable[extent - 1, ...] = chunk_arr
    else:
        chunktable[...] = chunk_arr

#----------------------------------------------------------------------------------
def expandChunk(chunk_shape, max_shape, typesize):
    """Extend the chunk shape until it is above the MIN target."""

    if chunk_shape is None:
        return None
    
    logging.debug(f"orig chunk_shape: {chunk_shape}")

    rank = len(chunk_shape)

    if rank != len(max_shape):
        raise ValueError("non-compatible arguments to expandChunk")

    if rank == 0:
        # scalar - can't be expanded
        return tuple(chunk_shape)
    
    chunk_shape = list(chunk_shape).copy()

    while True:
        chunk_size = np.prod(chunk_shape).item() * typesize
        if chunk_size >= MIN_CHUNK_SIZE:
            # this shape works
            break
        if chunk_size == 0:
            # can't do anything with zero-size chunks
            break

        extended = False

        for i in range(rank):
            # start from the low-order dimension
            dim = rank - i - 1 
            nextent = chunk_shape[dim]
            if nextent * 2 <= max_shape[dim]:
                chunk_shape[dim] = nextent * 2
                extended = True
                break

        if not extended:
            # unable to increase chunk_shape further
            break

    logging.debug(f"expandChunk - returning {chunk_shape}")

    return tuple(chunk_shape)


# ----------------------------------------------------------------------------------
def create_dataset(dobj, ctx):
    """create a dataset using the properties of the passed in h5py dataset.
    If successful, proceed to copy attributes and data.
    """
    chunks = None
    dset = None
    dset_preappend = None

    msg = f"create_dataset({dobj.name})"
    logging.info(msg)
    if ctx["verbose"]:
        print(msg)
    fout = ctx["fout"]

    if dobj.name in fout:
        dset = fout[dobj.name]
        logging.debug(f"{dobj.name} already exists")
        if ctx["no_clobber"]:
            msg = f"skipping creation of dataset {dobj.name} since already exist and no-clobber option is used"
            logging.info(msg)
            if ctx["verbose"]:
                print(msg)
            return
        if ctx["append"]:
            if ctx["extend_dim"]:
                msg = f"skipping creation of dataset {dobj.name} since already found"
                logging.info(msg)
                if ctx["verbose"]:
                    print(msg)
                return dset
            else:
                if len(dset.shape) == len(dobj.shape):
                    if dset.shape != dobj.shape:
                        msg = f"unable to append {dobj.name}: shape is not compatible"
                        logging.error(msg)    
                        if ctx["verbose"]:
                            print(msg)
                        return None
                    if len(dset.shape) == 0:
                        # don't try to extend scalar datasets (treat like attributes)
                        return dset

                    # add an extra dimension to append to
                    msg = f"re-creating {dobj.name} with extended dimension"
                    logging.info(msg)
                    if ctx["verbose"]:
                        print(msg)
                    dset_preappend = dset  # save to re-add data later
                    parent = dset.parent
                    obj_name = dobj.name.split("/")[-1]
                    logging.debug(f"removing link {dobj.name}")
                    del parent[obj_name]  # remove old link
                if len(dset.shape) == len(dobj.shape) + 1:
                    if dset.shape[1:] != dobj.shape:
                        msg = f"unable to append {dobj.name}: shape is not compatible"
                        logging.error(msg)
                        if not ctx["ignore_error"]:
                            raise IOError(msg)
                        return None
                    else:
                        # compatible shapes, can just return dset
                        return dset
        else:
            msg = f"unable to create dataset {dobj.name} already present"
            logging.error(msg)
            if not ctx["ignore_error"]:
                raise IOError(msg)
            return None
    else:
        if ctx["verbose"]:
            print(f"{dobj.name} not found")

    try:
        tgt_dtype = convert_dtype(dobj.dtype, ctx)
        if dset_preappend is not None:
            # add an extra unlimited dimension
            tgt_shape = [0,]
            tgt_maxshape = [None,]
        elif dobj.shape is None:
            tgt_shape = None
            tgt_maxshape = None
        else:
            tgt_shape = []
            tgt_maxshape = []

        if tgt_shape is None:
            rank = 0
        else:
            tgt_shape.extend(dobj.shape)
            tgt_maxshape.extend(dobj.maxshape)
            rank = len(tgt_shape)
        if rank > 0 and ctx["extend_dim"]:
            # set maxshape to unlimited for any dimension that is the extend_dim
            if dobj.name.split("/")[-1] == ctx["extend_dim"]:
                msg = f"setting {dobj.name} shape to unlimited"
                logging.info(msg)
                if ctx["verbose"]:
                    print(msg)
                if rank > 1:
                    msg = "expected extend dataset to be one-dimensional"
                    logging.warning(msg)
                    if ctx["verbose"]:
                        print(msg)
                for i in range(rank):
                    tgt_shape[i] = 0
                    tgt_maxshape[i] = None
            else:
                # check to see if any dimension scale refers to the extend dim
                for dim in range(len(dobj.dims)):
                    dimscales = dobj.dims[dim]
                    for i in range(len(dimscales)):
                        dimscale = dimscales[i]
                        if not dimscale:
                            continue
                        msg = f"dimscale for dim: {dim}: {dimscale}, type: {type(dimscale)}"
                        logging.debug(msg)
                        if dimscale.name.split("/")[-1] == ctx["extend_dim"]:
                            tgt_shape[dim] = 0
                            tgt_maxshape[dim] = None
                            msg = f"setting dimension {dim} of dataset {dobj.name} to unlimited"
                            logging.info(msg)
                            if ctx["verbose"]:
                                print(msg)
                            break

        kwargs = {"shape": tgt_shape, "maxshape": tgt_maxshape, "dtype": tgt_dtype}

        if (
            ctx["dataload"] in ("link", "fastlink")
            and not is_vlen(dobj.dtype)
            and dobj.shape is not None
            and len(dobj.shape) > 0
            and not is_compact(dobj)
            and np.prod(dobj.shape) > MIN_DSET_ELEMENTS_FOR_LINKING
        ):
            chunks = create_chunktable(dobj, tgt_shape, ctx)
            logging.info(f"using chunk layout: {chunks}")

        # use the source object layout if we are not using reference mapping
        if chunks is None and dobj.shape is not None and len(dobj.shape) > 0:
            # converting hsds dset with linked chunks to h5py dataset
            # just use the dims field of dobj.chunks as chunk shape
            chunks = get_chunk_dims(dobj)
        
        if chunks is not None and not is_h5py(fout):
            # expand chunk if too small            

            if dset_preappend is not None:
                # check to see if an extra dimension is needed for the chunk shape
                if isinstance(chunks, dict):
                    # chunktable is already adjusted
                    pass
                else:
                    new_chunks = [1,]
                    new_chunks.extend(chunks)
                    chunks = tuple(new_chunks)
            else:
                if isinstance(chunks, dict):
                    if "dims" in chunks:
                        chunk_dims = chunks["dims"]
                        if len(chunk_dims) == 1:
                            # currently hyperchunks only supported for 1d datasets
                            chunk_dims = expandChunk(chunk_dims, dobj.shape, dobj.dtype.itemsize)
                            logging.debug(f"expanded chunks: {chunk_dims}")
                            chunks["dims"] = chunk_dims 
                    else:
                        # contiguous or compact, using dataset shape
                        pass
                else:
                    # just a list with chunk shape
                    if len(chunks) == 1:
                        # currently hyperchunks only supported for 1d datasets
                        chunks = expandChunk(chunks, dobj.shape, dobj.dtype.itemsize)
                        logging.debug(f"expanded chunks: {chunks}")
                
            logging.debug(f"setting chunks kwargs to: {chunks}")
           
            kwargs["chunks"] = chunks

        if (
            dobj.shape is None
            or len(dobj.shape) == 0
            or (is_vlen(dobj.dtype) and is_h5py(fout))
        ):
            # don't use compression/chunks for scalar datasets
            # or vlen
            pass
        else:
            logging.debug(f"filter setup for {dobj.name}")
            if not ctx["ignorefilters"]:
                kwargs["compression"] = dobj.compression
                kwargs["compression_opts"] = dobj.compression_opts
                kwargs["shuffle"] = dobj.shuffle

            if ctx["default_compression"] is not None and not kwargs.get("compression"):
                kwargs["compression"] = ctx["default_compression"]
                kwargs["compression_opts"] = ctx["default_compression_opts"]
                if ctx["verbose"]:
                    print("applying default compression filter")
                
            # TBD: it would be better if HSDS could let us know what filters
            # are supported (like it does with compressors)
            # For now, just hard-ccreate_datasetcreate_datasetode fletcher32 and scaleoffset to be ignored
            if dobj.fletcher32:
                msg = f"fletcher32 filter used by dataset: {dobj.name} is not "
                msg += "supported by HSDS, this filter will not be used"
                logging.warning(msg)
                # kwargs["fletcher32"] = dobj.fletcher32
            if dobj.scaleoffset:
                msg = f"scaleoffset filter used by dataset: {dobj.name} is not "
                msg += "supported by HSDS, this filter will not be used"
                logging.warning(msg)

            if is_h5py(dobj) and not kwargs.get("compression"):
                # apply any custom filters as long as they are supported in HSDS
                for filter_id in dobj._filters:
                    filter_opts = dobj._filters[filter_id]
                    try:
                        filter_id = int(filter_id)
                    except ValueError:
                        msg = "unrecognized filter id: {filter_id} for {dobj.name}, ignoring"
                        logging.warning(msg)

                    if not isinstance(filter_id, int):
                        continue

                    if filter_id in H5Z_FILTER_MAP:
                        filter_name = H5Z_FILTER_MAP[filter_id]
                        if filter_name == "bitshuffle":
                            kwargs["shuffle"] = filter_name
                            logging.info(f"using bitshuffle on {dobj.name}")
                        else:
                            # supported non-standard compressor
                            kwargs["compression"] = filter_name
                            logging.info(f"using compressor: {filter_name} for {dobj.name}")
                            kwargs["compression_opts"] = filter_opts
                            logging.info(f"compression_opts: {filter_opts}")
                    else:
                        logging.warning(f"filter id {filter_id} for {dobj.name} not supported")

        # kwargs["scaleoffset"] = dobj.scaleoffset
        # setting the fillvalue is failing in some cases
        # see: https://github.com/HDFGroup/h5pyd/issues/119
        # don't set fill value for reference types
        fillvalue = get_fillvalue(dobj)
        if fillvalue is not None and not has_reference(tgt_dtype):
            logging.debug(f"got fillvalue: {fillvalue}")
            kwargs["fillvalue"] = fillvalue

        # finally, create the dataset  
        msg = f"creating dataset {dobj.name}, shape: {dobj.shape}, type: {tgt_dtype}"
        logging.info(msg)
        if ctx["verbose"]:
            print(msg) 
        dset = fout.create_dataset(dobj.name, **kwargs)

        msg = f"dataset created, uuid: {dset.id.id}, "
        msg += f"chunk_size: {str(dset.chunks)}, "
        msg += f"chunks: {chunks}"
        logging.info(msg)
        if ctx["verbose"]:
            print(msg)
        logging.debug(f"adding dataset id {dobj.id.id} to {dset}")
        srcid_desobj_map = ctx["srcid_desobj_map"]
        srcid_desobj_map[dobj.id.__hash__()] = dset
    except (IOError, TypeError, KeyError) as e:
        msg = f"ERROR: failed to create dataset: {e}"
        logging.error(msg)
        if not ctx["ignore_error"]:
            raise
        

    if dset_preappend is not None:
        write_dataset(dset_preappend, dset, ctx)
        # dset[0, ...] = dset_preappend[...]
    return dset

# ----------------------------------------------------------------------------------
def write_dataset(src, tgt, ctx):
    """write values from src dataset to target dataset."""
    msg = f"write_dataset src: {src.name} to tgt: {tgt.name}, shape: {src.shape}, type: {src.dtype}"
    logging.info(msg)
    if ctx["verbose"]:
        print(msg)
    if src.shape is None:
        # null space dataset
        msg = f"no data for null space dataset: {src.name}"
        logging.info(msg)
        if ctx["verbose"]:
            print(msg)
        return  # no data

    if len(src.shape) == 0:
        # scalar dataset
        x = src[()]
        msg = f"writing for scalar dataset: {src.name}"
        logging.info(msg)
        if ctx["verbose"]:
            print(msg)
        tgt[()] = x
        return

    if np.prod(src.shape) == 0:
        msg = f"no data for dataset with zero extent: {src.name}"
        logging.info(msg)
        if ctx["verbose"]:
            print(msg)
        return  # no data

    fillvalue = get_fillvalue(src)

    rank = len(src.shape)

    offset = [0,] * rank  # coordinate where we'll copy the source to

    if ctx["extend_dim"]:
        # resize dataset if a dimension is in the extended dimension
        if src.name.split("/")[-1] == ctx["extend_dim"]:
            if ctx["extend_offset"] is not None:
                offset[0] = ctx["extend_offset"]
            else:
                offset[0] = tgt.shape[0]
            new_extent = offset[0] + src.shape[0]
            if new_extent > tgt.shape[0]:
                msg = f"extending {tgt.name} shape to {new_extent}"
                logging.info(msg)
                if ctx["verbose"]:
                    print(msg)
                resize_dataset(tgt, new_extent, axis=0)
        else:
            # check to see if any dimension scale refers to the extend dim
            for dim in range(len(src.dims)):
                dimscales = src.dims[dim]
                for i in range(len(dimscales)):
                    dimscale = dimscales[i]
                    if not dimscale:
                        continue
                    msg = f"dimscale for dim: {dim}: {dimscale}, type: {type(dimscale)}"
                    logging.debug(msg)
                    if dimscale.name.split("/")[-1] == ctx["extend_dim"]:
                        if ctx["extend_offset"] is not None:
                            offset[dim] = ctx["extend_offset"]
                        else:
                            offset[dim] = tgt.shape[dim]
                        new_extent = offset[dim] + src.shape[dim]
                        if new_extent > tgt.shape[dim]:
                            msg = f"extending {tgt.name} shape to {new_extent} "
                            msg += f"for dimension: {dim}"
                            logging.info(msg)
                            if ctx["verbose"]:
                                print(msg)
                            resize_dataset(tgt, new_extent, axis=dim)
    elif len(tgt.shape) > len(src.shape):
        # append mode - extend first dimension by one
        new_extent = tgt.shape[0] + 1
        msg = f"extending first dimension to: {new_extent}"
        logging.info(msg)
        if ctx["verbose"]:
            print(msg)
        resize_dataset(tgt, new_extent, axis=0)

    if not is_h5py(tgt) and tgt.id.layout["class"] != "H5D_CHUNKED":
        # this is one of the ref layouts
        if tgt.id.layout["class"] == "H5D_CHUNKED_REF_INDIRECT":
            # don't write chunks, but update chunktable for chunk ref indirect
            update_chunktable(src, tgt, ctx)
        else:
            pass # skip chunkterator for link option
        return

    msg = f"iterating over chunks for {src.name}"
    logging.info(msg)
    if ctx["verbose"]:
        print(msg)
    try:
        if src.chunks is None:
            # contiguous dataset, fake an iterator by creating a list
            # with one slice
            slices = []
            for dim in range(rank):
                extent = src.shape[dim]
                s = slice(0, extent, 1)
                slices.append(s)
            slices = tuple(slices)
            if rank == 1:
                it = [slices[0],]
            else:
                it = [slices,]
        else:
            it = src.iter_chunks()

        logging.debug(f"src dtype: {src.dtype}")
        logging.debug(f"des dtype: {tgt.dtype}")

        for src_s in it:
            logging.debug(f"src selection: {src_s}")
            if rank == 1 and isinstance(src_s, slice):
                start = src_s.start + offset[0]
                stop = src_s.stop + offset[0]
                if len(tgt.shape) > rank:
                    tgt_s = []
                    n = tgt.shape[0] - 1
                    tgt_s.append(slice(n, n + 1, 1))
                    tgt_s.append(slice(start, stop, 1))
                    tgt_s = tuple(tgt_s)
                else:
                    tgt_s = slice(start, stop, 1)
            else:
                tgt_s = []
                if len(tgt.shape) > rank:
                    n = tgt.shape[0] - 1
                    tgt_s.append(slice(n, n + 1, 1))

                for dim in range(rank):
                    start = src_s[dim].start + offset[dim]
                    stop = src_s[dim].stop + offset[dim]
                    tgt_s.append(slice(start, stop, 1))
                tgt_s = tuple(tgt_s)
            logging.debug(f"tgt selection: {tgt_s}")

            arr = src[src_s]
            # don't write arr if it's all zeros (or the fillvalue if defined)
            empty_arr = np.zeros(arr.shape, dtype=arr.dtype)
            if fillvalue:
                empty_arr.fill(fillvalue)
            if np.array_equal(arr, empty_arr):
                msg = f"skipping chunk for slice: {src_s}"
            else:
                msg = f"writing dataset data for slice: {src_s}"
                arr = copy_array(arr, ctx)
                tgt[tgt_s] = arr
            logging.info(msg)
            if ctx["verbose"]:
                print(msg)
    except (IOError, TypeError) as e:
        msg = f"ERROR : failed to copy dataset data {src_s}: {e}"
        logging.error(msg)
        if not ctx["ignore_error"]:
            raise
    msg = f"done with dataload for {src.name}"
    logging.info(msg)
    if ctx["verbose"]:
        print(msg)
    logging.info("flush fout")


# write_dataset


def create_links(gsrc, gdes, ctx):
    # add soft and external links
    srcid_desobj_map = ctx["srcid_desobj_map"]
    msg = f"create_links: {gsrc.name}"
    logging.info(msg)
    if ctx["verbose"]:
        print(msg)
    for title in gsrc:
        msg = f"got link: {title}"
        if ctx["verbose"]:
            print(msg)
        logging.info(msg)
        lnk = gsrc.get(title, getlink=True)
        link_classname = lnk.__class__.__name__
        if link_classname == "HardLink":
            msg = f"Got hardlink: {title} gsrc: {gsrc} gdes: {gdes}"
            logging.debug(msg)
            if title not in gdes:
                msg = f"creating link {gdes} with title: {title}"
                if ctx["verbose"]:
                    print(msg)
                logging.info(msg)
                src_obj_id = gsrc[title].id
                src_obj_id_hash = src_obj_id.__hash__()
                logging.debug(f"got src_obj_id hash: {src_obj_id_hash}")
                if src_obj_id_hash in srcid_desobj_map:
                    des_obj = srcid_desobj_map[src_obj_id_hash]
                    logging.debug(f"creating hardlink to {des_obj.id.id}")
                    gdes[title] = des_obj
                else:
                    # TBD - in hdf5 1.10 it seems that two references to the same object
                    # can return different id's.  This will cause HDF5 files with
                    # multilinks to not load correctly
                    msg = f"could not find map item to src id: {src_obj_id_hash}"
                    logging.warn(msg)
                    if ctx["verbose"]:
                        print("WARNING: " + msg)
        elif link_classname == "SoftLink":
            msg = f"creating SoftLink({lnk.path}) with title: {title}"
            logging.info(msg)
            if ctx["verbose"]:
                print(msg)
            if is_h5py(gdes):
                soft_link = h5py.SoftLink(lnk.path)
            else:
                soft_link = h5pyd.SoftLink(lnk.path)
            gdes[title] = soft_link
        elif link_classname == "ExternalLink":
            msg = (
                f"creating ExternalLink({lnk.filename}, {lnk.path}) with title: {title}"
            )
            if ctx["verbose"]:
                print(msg)
            logging.info(msg)
            if is_h5py(gdes):
                ext_link = h5py.ExternalLink(lnk.filename, lnk.path)
            else:
                ext_link = h5pyd.ExternalLink(lnk.filename, lnk.path)
            gdes[title] = ext_link
        else:
            msg = f"Unexpected link type: {lnk.__class__.__name__}"
            logging.warning(msg)
            if ctx["verbose"]:
                print(msg)
    logging.info("flush fout")


# ----------------------------------------------------------------------------------
def create_group(gobj, ctx):
    msg = f"creating group {gobj.name}"
    logging.info(msg)
    if ctx["verbose"]:
        print(msg)
    fout = ctx["fout"]

    grp = None

    if gobj.name in fout:
        grp = fout[gobj.name]
        logging.debug(f"{gobj.name} already exists")
        if ctx["no_clobber"]:
            msg = f"skipping creation of group {gobj.name} since already exists and no-clobber mode is used"
            logging.info(msg)
            if ctx["verbose"]:
                print(msg)
            return
        
        if ctx["append"]:
            msg = f"skipping creation of group {gobj.name} since already found"
            logging.info(msg)
            if ctx["verbose"]:
                print(msg)
        else:
            msg = f"unable to create group {gobj.name}, already present"
            logging.error(msg)
            if not ctx["ignore_error"]:
                raise IOError(msg)
    else:
        if ctx["verbose"]:
            print(f"{gobj.name} not found")

        grp = fout.create_group(gobj.name)
        srcid_desobj_map = ctx["srcid_desobj_map"]
        msg = f"adding group id {gobj.id.id} to {grp} in srcid_desobj_map"
        logging.debug(msg)
        srcid_desobj_map[gobj.id.__hash__()] = grp

    return grp


# create_group


# -----------------------------------------------------------------------------
def create_datatype(obj, ctx):
    msg = f"creating datatype {obj.name}"
    logging.info(msg)
    if ctx["verbose"]:
        print(msg)
    fout = ctx["fout"]
    fout[obj.name] = obj.dtype
    srcid_desobj_map = ctx["srcid_desobj_map"]
    msg = f"adding datatype id {obj.id.id} to {fout[obj.name]} in srcid_desobj_map"
    logging.debug(msg)
    srcid_desobj_map[obj.id.__hash__()] = fout[obj.name]


# create_datatype

# ---------------------------------------------------------------------------------
def load_file(
    fin,
    fout,
    verbose=False,
    dataload="ingest",
    s3path=None,
    compression=None,
    compression_opts=None,
    ignorefilters=False,
    append=False,
    no_clobber=False,
    extend_dim=None,
    extend_offset=0,
    ignore_error=False,
):

    logging.info(f"input file: {fin.filename}")
    logging.info(f"output file: {fout.filename}")
    logging.info(f"dataload: {dataload}")
    if dataload != "ingest":
        if not dataload:
            logging.info("no data load")
        elif dataload in ("link", "fastlink"):
            if not s3path:
                logging.error("s3path expected to be set")
                sys.exit(1)
            logging.info(f"using s3path: {s3path}")
        else:
            logging.error(f"unexpected dataload value: {dataload}")
            sys.exit(1)

    # it would be nice to make a class out of these functions, but that
    # makes it hard to use visititems iterator.
    # instead, create a context object to pass arround common state
    ctx = {}
    ctx["fin"] = fin
    ctx["fout"] = fout
    ctx["verbose"] = verbose
    ctx["dataload"] = dataload  # ingest, link, or None
    ctx["default_compression"] = compression
    ctx["default_compression_opts"] = compression_opts
    ctx["ignorefilters"] = ignorefilters
    ctx["s3path"] = s3path
    ctx["append"] = append
    ctx["no_clobber"] = no_clobber
    ctx["extend_dim"] = extend_dim
    ctx["extend_offset"] = extend_offset
    ctx["srcid_desobj_map"] = {}
    ctx["ignore_error"] = ignore_error

    def copy_attribute_helper(name, obj):
        logging.info(f"copy attribute - name: {name}  obj: {obj.name}")
        tgt = fout[name]
        for a in obj.attrs:
            copy_attribute(tgt, a, obj, ctx)

    def object_create_helper(name, obj):
        logging.info(f"object create helper - name: {name} obj: {obj.name}")
        class_name = obj.__class__.__name__
        if class_name in ("Dataset", "Table"):
            create_dataset(obj, ctx)
        elif class_name == "Group":
            create_group(obj, ctx)
        elif class_name == "Datatype":
            create_datatype(obj, ctx)

    def object_link_helper(name, obj):
        class_name = obj.__class__.__name__
        logging.info(f"object_link_helper for object: {obj.name}")
        if class_name == "Group":
            # create any soft/external links
            fout = ctx["fout"]
            grp = fout[name]
            create_links(obj, grp, ctx)

    def object_copy_helper(name, obj):
        class_name = obj.__class__.__name__
        logging.debug(f"object_copy_helper for object: {obj.name}")

        if class_name in ("Dataset", "Table"):
            logging.debug(f"calling write_dataset for dataset: {obj.name}")
            tgt = fout[obj.name]
            write_dataset(obj, tgt, ctx)
        elif class_name == "Group":
            logging.debug(f"skip copy for group: {obj.name}")
        elif class_name == "Datatype":
            logging.debug(f"skip copy for datatype: {obj.name}")
        else:
            logging.error(f"no handler for object class: {type(obj)}")

    # build a rough map of the file using the internal function above
    # copy over any attributes
    # create soft/external links (and hardlinks not already created)
    logging.info("creating target objects and attributes")

    # build a rough map of the file using the internal function above
    logging.info("creating target objects")
    fin.visititems(object_create_helper)

    # copy over any attributes
    logging.info("creating target attributes")
    fin.visititems(copy_attribute_helper)

    # create soft/external links (and hardlinks not already created)
    create_links(fin, fout, ctx)  # create root soft/external links
    fin.visititems(object_link_helper)

    if dataload == "ingest" or dataload == "link":
        # copy dataset data
        logging.info("copying dataset data")
        fin.visititems(object_copy_helper)
    else:
        logging.info("skipping dataset data copy (dataload is None)")

    # create any root attributes
    for ga in fin.attrs:
        copy_attribute(fout, ga, fin, ctx)

    # Fully flush the h5py handle.
    fout.close()

    # close up the source domain, see reason(s) for this below
    fin.close()
    msg = "load_file complete"
    logging.info(msg)
    if verbose:
        print(msg)

    return 0
    # load_file



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HSDS (HDF5 Scalable Data Service), Libraries and      #
# Utilities.  The full HSDS copyright notice, including                      #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

import sys
import logging
from datetime import datetime
import h5pyd

if __name__ == "__main__":
    from config import Config
else:
    from .config import Config

#
# Print objects in a domain in the style of the hsls utilitiy
#

cfg = Config()

#
# log error and abort app
#
def abort(msg):
    logging.error(msg)
    if cfg["logfile"]:
        # write to stderr if we are output logs to a file
        sys.stderr.write(msg + "\n")
    logging.error("exiting program with return code -1")
    sys.exit(-1)

#
# Usage
#
def usage():
    option_names = cfg.get_names()
    cmd = cfg.get_cmd()
    print("Usage:\n")
    print(f"    {cmd} [ OPTIONS ]  domain")
    print(f"    {cmd} [ OPTIONS ]  folder")
    print("")
    print("Description:")
    print("    Get domain stats for domain or folder")
    print("       domain: HSDS domain (absolute path with or without 'hdf5:// prefix)")
    print("       folder: HSDS folder (path as above ending in '/')")
    print("")
    
    print("Options:")
    for name in option_names:
        help_msg = cfg.get_help_message(name)
        if help_msg:
            print(f"    {help_msg}")  
    print("")
    print("examples:")
    print(f"   {cmd} -e http://hsdshdflab.hdfgroup.org")
    print(f"   {cmd} -e http://hsdshdflab.hdfgroup.org /shared/tall.h5")
    print(cfg.get_see_also(cmd))
    print("")
    sys.exit()


def format_size(n):
    if n is None or n == " ":
        return " " * 8
    symbol = " "
    if not cfg["human_readable"]:
        return str(n)
    # convert to common storage unit
    for s in ("B", "K", "M", "G", "T"):
        if n < 1024:
            symbol = s
            break
        n /= 1024
    if symbol == "B":
        return "{:}B".format(n)
    else:
        return "{:.1f}{}".format(n, symbol)


def getDomainInfo(domain, cfg):
    """get info about the domain and print"""
    username = cfg["hs_username"]
    password = cfg["hs_password"]
    endpoint = cfg["hs_endpoint"]
    bucket = cfg["hs_bucket"]
    if "rescan" in cfg and cfg["rescan"]:
        mode = "r+"  # need write intent
    else:
        mode = "r"

    if domain.endswith("/"):
        is_folder = True
    else:
        is_folder = False

    try:
        if is_folder:
            f = h5pyd.Folder(
                domain,
                mode=mode,
                endpoint=endpoint,
                username=username,
                password=password,
                bucket=bucket,
                use_cache=True,
            )
        else:
            f = h5pyd.File(
                domain,
                mode=mode,
                endpoint=endpoint,
                username=username,
                password=password,
                bucket=bucket,
                use_cache=False,
            )
    except IOError as oe:
        if oe.errno in (404, 410):  # Not Found
            abort(f"domain: {domain} not found")
        elif oe.errno == 401:  # Unauthorized
            abort("Authorization failure")
        elif oe.errno == 403:  # Forbidden
            abort("Not allowed")
        else:
            abort(f"Unexpected error: {oe}")

    timestamp = datetime.fromtimestamp(int(f.modified))
    if not is_folder and f.last_scan:
        last_scan = datetime.fromtimestamp(int(f.last_scan))
    else:
        last_scan = None

    if is_folder:
        print(f"folder: {domain}")
        print(f"    owner:           {f.owner}")
        print(f"    last modified:   {timestamp}")
    else:
        if "rescan" in cfg and cfg["rescan"]:
            f.run_scan()

        # report HDF objects (groups, datasets, and named datatypes) vs. allocated chunks
        num_objects = f.num_groups + f.num_datatypes + f.num_datasets
        if f.num_chunks > 0:
            num_chunks = f.num_chunks
        else:
            # older storeinfo format doesn't have num_chunks, so calculate
            num_chunks = f.num_objects - num_objects

        print(f"domain: {domain}")
        print(f"    owner:           {f.owner}")
        print(f"    id:              {f.id.id}")
        print(f"    last modified:   {timestamp}")
        if last_scan:
            print(f"    last scan:       {last_scan}")
        if f.md5_sum:
            print(f"    md5 sum:         {f.md5_sum}")
        print(f"    total_size:      {format_size(f.total_size)}")
        print(f"    allocated_bytes: {format_size(f.allocated_bytes)}")
        if f.metadata_bytes:
            print(f"    metadata_bytes:  {format_size(f.metadata_bytes)}")
        if f.linked_bytes:
            print(f"    linked_bytes:    {format_size(f.linked_bytes)}")
        print(f"    num objects:     {num_objects}")
        print(f"    num chunks:      {num_chunks}")
        if f.num_linked_chunks:
            print(f"    linked chunks:   {f.num_linked_chunks}")

    f.close()
#
# Main
#
def main():
    domains = []

    cfg.setitem("human_readable", False, flags=["-H", "--human-readable"], help="print human readable sizes (e.g. 123M)")
    cfg.setitem("rescan", False, flags=["--rescan",], help="refresh domain stats (for use when domain is provided)")
    cfg.setitem("help", False, flags=["-h", "--help"], help="this message")

    try:
        domains = cfg.set_cmd_flags(sys.argv[1:])
    except ValueError as ve:
        print(ve)
        usage()

    # setup logging
    logfname = cfg["logfile"]
    loglevel = cfg.get_loglevel()
    logging.basicConfig(filename=logfname, format='%(levelname)s %(asctime)s %(message)s', level=loglevel)
    logging.debug(f"set log_level to {loglevel}")

    if not domains:
        abort("no domain provided!")

    for domain in domains:
        getDomainInfo(domain, cfg)


if __name__ == "__main__":
    main()



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HSDS (HDF5 Scalable Data Service), Libraries and      #
# Utilities.  The full HSDS copyright notice, including                      #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

import sys
import logging
import os.path as op
import h5pyd

if __name__ == "__main__":
    from config import Config
else:
    from .config import Config

cfg = Config()

#----------------------------------------------------------------------------------
def usage():
    option_names = cfg.get_names()
    cmd = cfg.get_cmd()
    print("Usage:\n")
    print(f"    {cmd} [ OPTIONS ]  source_domain  des_domain")
    print(f"    {cmd} [ OPTIONS ]  source_domain  folder")
    print("")
    print("Description:")
    print("    Move domain from one location to another")
    print("       source_domain: domain to be moved ")
    print("       des_domain: destination domain")
    print("       folder: destination folder (Unix style ending in '/')")
    print("")
    print("Options:")
    for name in option_names:
        help_msg = cfg.get_help_message(name)
        if help_msg:
            print(f"    {help_msg}")  
    print("")
    print("Examples:")
    print(f"     {cmd} /home/myfolder/file1.h5  /home/myfolder/file2.h5")
    print(f"     {cmd} /home/myfolder/file2.h5  /home/anotherfolder/")
    print("")
    print(cfg.get_see_also(cmd))
    print("")
    sys.exit(-1)
#end print_usage

#----------------------------------------------------------------------------------
def print_config_example():
    print("# default")
    print("hs_username = <username>")
    print("hs_password = <passwd>")
    print("hs_endpoint = http://hsdshdflab.hdfgroup.org")
#print_config_example

#----------------------------------------------------------------------------------

def getFolder(domain, mode="r"):
    username = cfg["hs_username"]
    password = cfg["hs_password"]
    endpoint = cfg["hs_endpoint"]
    bucket = cfg["hs_bucket"]
    dir = h5pyd.Folder(domain, endpoint=endpoint, username=username,
                      password=password, bucket=bucket, mode=mode)
    return dir


def getFile(domain, mode="r"):
    username = cfg["hs_username"]
    password = cfg["hs_password"]
    endpoint = cfg["hs_endpoint"]
    bucket = cfg["hs_bucket"]
    fh = h5pyd.File(domain, mode=mode, endpoint=endpoint, username=username,
                   password=password, bucket=bucket, use_cache=True)
    return fh

def createFile(domain, linked_domain=None, no_clobber=False):
    #print("createFile", domain)
    username = cfg["hs_username"]
    password = cfg["hs_password"]
    endpoint = cfg["hs_endpoint"]
    bucket = cfg["hs_bucket"]
    owner = None
    if no_clobber:
        mode= "x"
    else:
        mode="w"
    if "hs_owner" in cfg:
        owner=cfg["hs_owner"]
    fh = h5pyd.File(domain, mode=mode, endpoint=endpoint, username=username, password=password, bucket=bucket, owner=owner, linked_domain=linked_domain)
    return fh


def deleteDomain(domain, keep_root=False):

    # get handle to parent folder
    if domain.endswith('/'):
        path = domain[:-1]
    else:
        path = domain
    parent_domain = op.dirname(path)
    base_name = op.basename(path)

    if len(parent_domain) < 2:
        sys.exit("can't get parent domain")

    if not parent_domain.endswith('/'):
        parent_domain += '/'
    try:
        hparent = getFolder(parent_domain, mode='a')
    except IOError as oe:
        if oe.errno == 404:   # Not Found
            sys.exit(f"Parent domain: {parent_domain} not found")
        elif oe.errno == 401:  # Unauthorized
            sys.exit("Authorization failure")
        elif oe.errno == 403:  # Forbidden
            sys.exit("Not allowed")
        else:
            sys.exit(f"Unexpected error: {oe}")

    if base_name not in hparent:
        # note - this may happen if the domain was recently created and not
        # yet synced to S3
        sys.exit(f"domain: {domain} not found")

    # delete the domain
    hparent.delete_item(base_name, keep_root=keep_root)
    if cfg["verbose"]:
        if domain.endswith('/'):
            print(f"Folder: {domain} deleted")
        else:
            print(f"Domain: {domain} deleted")

def main():

    cfg.setitem("no_clobber", False, flags=["-n", "--no-clobber"],  help="do not overwrite any domains")
    cfg.setitem("hs_owner", None, flags=["-o", "--owner"], choices=["OWNER",], help="set owner (must be run as an admin user)")
    cfg.setitem("help", False, flags=["-h", "--help"], help="this message")

    try:
        domains = cfg.set_cmd_flags(sys.argv[1:])
    except ValueError as ve:
        print(ve)
        usage()

    if len(domains) < 2:
        # need at least source and target
        usage()

    # setup logging
    logfname = cfg["logfile"]
    loglevel = cfg.get_loglevel()
    logging.basicConfig(filename=logfname, format='%(levelname)s %(asctime)s %(message)s', level=loglevel)
    logging.debug(f"set log_level to {loglevel}")

    src_domain = domains[0]
    des_domain = domains[1]

    if src_domain[0] != '/' or des_domain[0] != '/':
        print("absolute paths must be used")
        usage()

    if src_domain[-1] == '/':
        print("folder can not be used for src domain")

    if des_domain[-1] == '/':
        # add on the filename from source to folder path
        des_domain += op.basename(src_domain)

    logging.info(f"source domain: {src_domain}")
    logging.info(f"target domain: {des_domain}")

    # get root id of source file
    try:
        fin = getFile(src_domain)
    except IOError as oe:
        # this will fail if we try to open a folder
        msg = f"Error: {oe.errno} getting domain: {src_domain}"
        logging.error(msg)
        sys.exit(str(oe))

    logging.info(f"src root id: {fin.id.id}")
    fin.close()

    # create a new file using the src domain for the root group
    no_clobber = cfg["no_clobber"]
    try:
        fout = createFile(des_domain, linked_domain=src_domain, no_clobber=no_clobber)
    except IOError as oe:
        msg = f"Error: {oe.errno} creating domain: {des_domain}"
        logging.error(msg)
        sys.exit(str(oe))

    cfg.print(f"{src_domain} copied to {des_domain}")
    logging.info(f"des root id: {fout.id.id}")

    try:
        deleteDomain(src_domain, keep_root=True)
        logging.info(f"domain: {src_domain} removed")
    except IOError as oe:
        msg = f"Error: {oe.errno} removing source domain: {src_domain}"
        logging.error(msg)
        sys.exit(str(oe))

# __main__
if __name__ == "__main__":
    main()
